/****************************************************************************
 * Copyright (C) 2016 by University of Warsaw                               *
 *                                                                          *
 * This file is part of IntuitionTC.                                        *
 *                                                                          *
 ****************************************************************************/
/**
 * @file alloc.h
 * @brief Allocation, creation and copying of basic structures; simple freeing
 *        operations.
 * @author Aleksy Schubert
 * @author Maciej Zielenkiewicz
 * @author Marcin Benke
 */
#ifndef _ALLOC_H
#define _ALLOC_H  1
#include "checker.h"





/*@
  @ behavior allocation:
  @   assumes is_allocable{Pre}((size_t)sizeof(Term),(size_t)1);
  @   assigns __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   ensures allocation_footprint{Pre,Post}((size_t)(sizeof(Term)),(size_t)1);
  @   ensures is_allocated_size{Post}((char*)\result,(size_t)(sizeof(Term)));
  @   ensures \valid((char*)\result+(0..sizeof(Term)-1));
  @   ensures \result!=\null;
  @   ensures \forall char* p; \result!=\null && (is_allocated{Pre}(p) ==> p != (char*)\result);
  @   ensures \forall char* p, size_t n; is_allocated_size{Pre}(p,n) ==> is_allocated_size{Post}(p,n);
  @   ensures \forall size_t i; 0 <= i <= \at(__num_allocated,Post)-1 ==> \at(__allocated[i],Post) == \at(__allocated[i],Pre);
  @   ensures __allocated[__num_allocated-1] == (char*)\result;
  @   ensures \separated(__allocated+(0..MAX_ALLOCS-1),(char*)\result+(0..sizeof(Term)-1));
  @   ensures \forall char** a, size_t s; is_allocated_size{Pre}((char*)a, s) ==> 
  @                     \separated((char*)\result+(0..sizeof(Term)-1),a+(0..s-1));
  @ behavior no_allocation:
  @   assumes !is_allocable{Pre}((size_t)sizeof(Term),(size_t)1);
  @   assigns \nothing;
  @   ensures \result==\null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Term* allocTerm(); 

/*@
  @   requires  is_allocated{Pre}((char*)ptr);
  @   assigns  __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   frees ptr;
  @   ensures !is_allocated{Post}((char*)ptr);
  @   ensures old_allocated{Pre,Post}((char *)ptr);
  @   ensures \forall size_t s; is_allocated_size{Pre}((char*)ptr,s) ==> \at(__size_allocated,Post) == \at(__size_allocated,Pre) - s;
  @   ensures allocation_footprint{Post,Pre}((size_t)(\block_length{Pre}((char*)ptr)), (size_t)1);
  @*/
void freeTerm(Term* ptr); 
void freeTermZ(Term**);

/*@
  @ behavior allocation:
  @   assumes is_allocable{Pre}((size_t)sizeof(Term),(size_t)1);
  @   assigns __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   ensures allocation_footprint{Pre,Post}((size_t)(sizeof(Term)),(size_t)1);
  @   ensures is_allocated_size{Post}((char*)\result,(size_t)(sizeof(Term)));
  @   ensures \valid((char*)\result+(0..sizeof(Term)-1));
  @   ensures \result!=\null;
  @   ensures \forall char* p; \result!=\null && (is_allocated{Pre}(p) ==> p != (char*)\result);
  @   ensures \forall char* p, size_t n; is_allocated_size{Pre}(p,n) ==> is_allocated_size{Post}(p,n);
  @   ensures \forall size_t i; 0 <= i <= \at(__num_allocated,Post)-1 ==> \at(__allocated[i],Post) == \at(__allocated[i],Pre);
  @   ensures __allocated[__num_allocated-1] == (char*)\result;
  @   ensures \separated(__allocated+(0..MAX_ALLOCS-1),(char*)\result+(0..sizeof(Term)-1));
  @   ensures \forall char** a, size_t s; is_allocated_size{Pre}((char*)a, s) ==> 
  @                     \separated((char*)\result+(0..sizeof(Term)-1),a+(0..s-1));
  @   ensures \forall char* a, size_t s; is_allocated_size{Pre}(a, s) ==> 
  @                     \separated((char*)\result+(0..sizeof(Term)-1),a+(0..s-1));
  @   ensures isShallowCopyOfTerm{Post,Pre}(*\result,arg);
  @ behavior no_allocation:
  @   assumes !is_allocable{Pre}((size_t)sizeof(Term),(size_t)1);
  @   assigns \nothing;
  @   ensures \result==\null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Term* makeTerm(const Term arg); /* shallow copy */

/*@ requires valid_Term_ptr(arg);
  @ behavior allocation:
  @   assumes is_allocable{Pre}((size_t)sizeof(Term),(size_t)1);
  @   assigns __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   ensures allocation_footprint{Pre,Post}((size_t)(sizeof(Term)),(size_t)1);
  @   ensures is_allocated_size{Post}((char*)\result,(size_t)(sizeof(Term)));
  @   ensures \valid((char*)\result+(0..sizeof(Term)-1));
  @   ensures \result!=\null;
  @   ensures isShallowCopyOfTerm{Post,Pre}(*\result,*arg);
  @ behavior no_allocation:
  @   assumes !is_allocable{Pre}((size_t)sizeof(Term),(size_t)1);
  @   assigns \nothing;
  @   ensures \result==\null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Term* copyTerm(const Term* arg); /* shallow copy */
Term* copyTermDeep(const Term*); /* deep copy */

/*@
  @ behavior allocation:
  @   assumes is_allocable{Pre}((size_t)sizeof(Type),(size_t)1);
  @   assigns __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   ensures allocation_footprint{Pre,Post}((size_t)(sizeof(Type)),(size_t)1);
  @   ensures is_allocated_size{Post}((char*)\result,(size_t)(sizeof(Type)));
  @   ensures \valid((char*)\result+(0..sizeof(Type)-1));
  @   ensures \result!=\null;
  @   ensures \forall char* p; \result!=\null && (is_allocated{Pre}(p) ==> p != (char*)\result);
  @   ensures \forall char* p, size_t n; is_allocated_size{Pre}(p,n) ==> is_allocated_size{Post}(p,n);
  @   ensures \forall size_t i; 0 <= i <= \at(__num_allocated,Post)-1 ==> \at(__allocated[i],Post) == \at(__allocated[i],Pre);
  @   ensures __allocated[__num_allocated-1] == (char*)\result;
  @   ensures \separated(__allocated+(0..MAX_ALLOCS-1),(char*)\result+(0..sizeof(Type)-1));
  @   ensures \forall char** a, size_t s; is_allocated_size{Pre}((char*)a, s) ==> 
  @                     \separated((char*)\result+(0..sizeof(Type)-1),a+(0..s-1));
  @ behavior no_allocation:
  @   assumes !is_allocable{Pre}((size_t)sizeof(Type),(size_t)1);
  @   assigns \nothing;
  @   ensures \result==\null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Type* allocType(); 

/*@
  @   requires  is_allocated{Pre}((char*)ptr);
  @   assigns  __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   frees ptr;
  @   ensures !is_allocated{Post}((char*)ptr);
  @   ensures old_allocated{Pre,Post}((char *)ptr);
  @   ensures \forall size_t s; is_allocated_size{Pre}((char*)ptr,s) ==> \at(__size_allocated,Post) == \at(__size_allocated,Pre) - s;
  @   ensures allocation_footprint{Post,Pre}((size_t)(\block_length{Pre}((char*)ptr)), (size_t)1);
  @*/
void freeType(Type* ptr); 
void freeTypeZ(Type**);

/*@
  @ behavior allocation:
  @   assumes is_allocable{Pre}((size_t)sizeof(Type),(size_t)1);
  @   assigns __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   ensures allocation_footprint{Pre,Post}((size_t)(sizeof(Type)),(size_t)1);
  @   ensures is_allocated_size{Post}((char*)\result,(size_t)(sizeof(Type)));
  @   ensures \valid((char*)\result+(0..sizeof(Type)-1));
  @   ensures \result!=\null;
  @   ensures \forall char* p; \result!=\null && (is_allocated{Pre}(p) ==> p != (char*)\result);
  @   ensures \forall char* p, size_t n; is_allocated_size{Pre}(p,n) ==> is_allocated_size{Post}(p,n);
  @   ensures \forall size_t i; 0 <= i <= \at(__num_allocated,Post)-1 ==> \at(__allocated[i],Post) == \at(__allocated[i],Pre);
  @   ensures __allocated[__num_allocated-1] == (char*)\result;
  @   ensures \separated(__allocated+(0..MAX_ALLOCS-1),(char*)\result+(0..sizeof(Type)-1));
  @   ensures \forall char** a, size_t s; is_allocated_size{Pre}((char*)a, s) ==> 
  @                     \separated((char*)\result+(0..sizeof(Type)-1),a+(0..s-1));
  @   ensures \forall char* a, size_t s; is_allocated_size{Pre}(a, s) ==> 
  @                     \separated((char*)\result+(0..sizeof(Type)-1),a+(0..s-1));
  @   ensures isShallowCopyOfType{Post,Pre}(*\result,arg);
  @ behavior no_allocation:
  @   assumes !is_allocable{Pre}((size_t)sizeof(Type),(size_t)1);
  @   assigns \nothing;
  @   ensures \result==\null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Type* makeType(const Type arg); /* shallow copy */

/*@ requires valid_Type_ptr(arg);
  @ behavior allocation:
  @   assumes is_allocable{Pre}((size_t)sizeof(Type),(size_t)1);
  @   assigns __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   ensures allocation_footprint{Pre,Post}((size_t)(sizeof(Type)),(size_t)1);
  @   ensures is_allocated_size{Post}((char*)\result,(size_t)(sizeof(Type)));
  @   ensures \valid((char*)\result+(0..sizeof(Type)-1));
  @   ensures \result!=\null;
  @   ensures isShallowCopyOfType{Post,Pre}(*\result,*arg);
  @ behavior no_allocation:
  @   assumes !is_allocable{Pre}((size_t)sizeof(Type),(size_t)1);
  @   assigns \nothing;
  @   ensures \result==\null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Type* copyType(const Type* arg); /* shallow copy */
Type* copyTypeDeep(const Type*); /* deep copy */

/*@
  @ behavior allocation:
  @   assumes is_allocable{Pre}((size_t)sizeof(Kind),(size_t)1);
  @   assigns __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   ensures allocation_footprint{Pre,Post}((size_t)(sizeof(Kind)),(size_t)1);
  @   ensures is_allocated_size{Post}((char*)\result,(size_t)(sizeof(Kind)));
  @   ensures \valid((char*)\result+(0..sizeof(Kind)-1));
  @   ensures \result!=\null;
  @   ensures \forall char* p; \result!=\null && (is_allocated{Pre}(p) ==> p != (char*)\result);
  @   ensures \forall char* p, size_t n; is_allocated_size{Pre}(p,n) ==> is_allocated_size{Post}(p,n);
  @   ensures \forall size_t i; 0 <= i <= \at(__num_allocated,Post)-1 ==> \at(__allocated[i],Post) == \at(__allocated[i],Pre);
  @   ensures __allocated[__num_allocated-1] == (char*)\result;
  @   ensures \separated(__allocated+(0..MAX_ALLOCS-1),(char*)\result+(0..sizeof(Kind)-1));
  @   ensures \forall char** a, size_t s; is_allocated_size{Pre}((char*)a, s) ==> 
  @                     \separated((char*)\result+(0..sizeof(Kind)-1),a+(0..s-1));
  @ behavior no_allocation:
  @   assumes !is_allocable{Pre}((size_t)sizeof(Kind),(size_t)1);
  @   assigns \nothing;
  @   ensures \result==\null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Kind* allocKind(); 

/*@
  @   requires  is_allocated{Pre}((char*)ptr);
  @   assigns  __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   frees ptr;
  @   ensures !is_allocated{Post}((char*)ptr);
  @   ensures old_allocated{Pre,Post}((char *)ptr);
  @   ensures \forall size_t s; is_allocated_size{Pre}((char*)ptr,s) ==> \at(__size_allocated,Post) == \at(__size_allocated,Pre) - s;
  @   ensures allocation_footprint{Post,Pre}((size_t)(\block_length{Pre}((char*)ptr)), (size_t)1);
  @*/
void freeKind(Kind* ptr); 
void freeKindZ(Kind**);

/*@
  @ behavior allocation:
  @   assumes is_allocable{Pre}((size_t)sizeof(Kind),(size_t)1);
  @   assigns __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   ensures allocation_footprint{Pre,Post}((size_t)(sizeof(Kind)),(size_t)1);
  @   ensures is_allocated_size{Post}((char*)\result,(size_t)(sizeof(Kind)));
  @   ensures \valid((char*)\result+(0..sizeof(Kind)-1));
  @   ensures \result!=\null;
  @   ensures \forall char* p; \result!=\null && (is_allocated{Pre}(p) ==> p != (char*)\result);
  @   ensures \forall char* p, size_t n; is_allocated_size{Pre}(p,n) ==> is_allocated_size{Post}(p,n);
  @   ensures \forall size_t i; 0 <= i <= \at(__num_allocated,Post)-1 ==> \at(__allocated[i],Post) == \at(__allocated[i],Pre);
  @   ensures __allocated[__num_allocated-1] == (char*)\result;
  @   ensures \separated(__allocated+(0..MAX_ALLOCS-1),(char*)\result+(0..sizeof(Kind)-1));
  @   ensures \forall char** a, size_t s; is_allocated_size{Pre}((char*)a, s) ==> 
  @                     \separated((char*)\result+(0..sizeof(Kind)-1),a+(0..s-1));
  @   ensures \forall char* a, size_t s; is_allocated_size{Pre}(a, s) ==> 
  @                     \separated((char*)\result+(0..sizeof(Kind)-1),a+(0..s-1));
  @   ensures isShallowCopyOfKind{Post,Pre}(*\result,arg);
  @ behavior no_allocation:
  @   assumes !is_allocable{Pre}((size_t)sizeof(Kind),(size_t)1);
  @   assigns \nothing;
  @   ensures \result==\null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Kind* makeKind(const Kind arg); /* shallow copy */

/*@ requires valid_Kind_ptr(arg);
  @ behavior allocation:
  @   assumes is_allocable{Pre}((size_t)sizeof(Kind),(size_t)1);
  @   assigns __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   ensures allocation_footprint{Pre,Post}((size_t)(sizeof(Kind)),(size_t)1);
  @   ensures is_allocated_size{Post}((char*)\result,(size_t)(sizeof(Kind)));
  @   ensures \valid((char*)\result+(0..sizeof(Kind)-1));
  @   ensures \result!=\null;
  @   ensures isShallowCopyOfKind{Post,Pre}(*\result,*arg);
  @ behavior no_allocation:
  @   assumes !is_allocable{Pre}((size_t)sizeof(Kind),(size_t)1);
  @   assigns \nothing;
  @   ensures \result==\null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Kind* copyKind(const Kind* arg); /* shallow copy */
Kind* copyKindDeep(const Kind*); /* deep copy */

/*@
  @ behavior allocation:
  @   assumes is_allocable{Pre}((size_t)sizeof(Any),(size_t)1);
  @   assigns __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   ensures allocation_footprint{Pre,Post}((size_t)(sizeof(Any)),(size_t)1);
  @   ensures is_allocated_size{Post}((char*)\result,(size_t)(sizeof(Any)));
  @   ensures \valid((char*)\result+(0..sizeof(Any)-1));
  @   ensures \result!=\null;
  @   ensures \forall char* p; \result!=\null && (is_allocated{Pre}(p) ==> p != (char*)\result);
  @   ensures \forall char* p, size_t n; is_allocated_size{Pre}(p,n) ==> is_allocated_size{Post}(p,n);
  @   ensures \forall size_t i; 0 <= i <= \at(__num_allocated,Post)-1 ==> \at(__allocated[i],Post) == \at(__allocated[i],Pre);
  @   ensures __allocated[__num_allocated-1] == (char*)\result;
  @   ensures \separated(__allocated+(0..MAX_ALLOCS-1),(char*)\result+(0..sizeof(Any)-1));
  @   ensures \forall char** a, size_t s; is_allocated_size{Pre}((char*)a, s) ==> 
  @                     \separated((char*)\result+(0..sizeof(Any)-1),a+(0..s-1));
  @ behavior no_allocation:
  @   assumes !is_allocable{Pre}((size_t)sizeof(Any),(size_t)1);
  @   assigns \nothing;
  @   ensures \result==\null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Any* allocAny(); 

/*@
  @   requires  is_allocated{Pre}((char*)ptr);
  @   assigns  __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   frees ptr;
  @   ensures !is_allocated{Post}((char*)ptr);
  @   ensures old_allocated{Pre,Post}((char *)ptr);
  @   ensures \forall size_t s; is_allocated_size{Pre}((char*)ptr,s) ==> \at(__size_allocated,Post) == \at(__size_allocated,Pre) - s;
  @   ensures allocation_footprint{Post,Pre}((size_t)(\block_length{Pre}((char*)ptr)), (size_t)1);
  @*/
void freeAny(Any* ptr); 
void freeAnyZ(Any**);

/*@
  @ behavior allocation:
  @   assumes is_allocable{Pre}((size_t)sizeof(Any),(size_t)1);
  @   assigns __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   ensures allocation_footprint{Pre,Post}((size_t)(sizeof(Any)),(size_t)1);
  @   ensures is_allocated_size{Post}((char*)\result,(size_t)(sizeof(Any)));
  @   ensures \valid((char*)\result+(0..sizeof(Any)-1));
  @   ensures \result!=\null;
  @   ensures \forall char* p; \result!=\null && (is_allocated{Pre}(p) ==> p != (char*)\result);
  @   ensures \forall char* p, size_t n; is_allocated_size{Pre}(p,n) ==> is_allocated_size{Post}(p,n);
  @   ensures \forall size_t i; 0 <= i <= \at(__num_allocated,Post)-1 ==> \at(__allocated[i],Post) == \at(__allocated[i],Pre);
  @   ensures __allocated[__num_allocated-1] == (char*)\result;
  @   ensures \separated(__allocated+(0..MAX_ALLOCS-1),(char*)\result+(0..sizeof(Any)-1));
  @   ensures \forall char** a, size_t s; is_allocated_size{Pre}((char*)a, s) ==> 
  @                     \separated((char*)\result+(0..sizeof(Any)-1),a+(0..s-1));
  @   ensures \forall char* a, size_t s; is_allocated_size{Pre}(a, s) ==> 
  @                     \separated((char*)\result+(0..sizeof(Any)-1),a+(0..s-1));
  @   ensures isShallowCopyOfAny{Post,Pre}(*\result,arg);
  @ behavior no_allocation:
  @   assumes !is_allocable{Pre}((size_t)sizeof(Any),(size_t)1);
  @   assigns \nothing;
  @   ensures \result==\null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Any* makeAny(const Any arg); /* shallow copy */

/*@ requires valid_Any_ptr(arg);
  @ behavior allocation:
  @   assumes is_allocable{Pre}((size_t)sizeof(Any),(size_t)1);
  @   assigns __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   ensures allocation_footprint{Pre,Post}((size_t)(sizeof(Any)),(size_t)1);
  @   ensures is_allocated_size{Post}((char*)\result,(size_t)(sizeof(Any)));
  @   ensures \valid((char*)\result+(0..sizeof(Any)-1));
  @   ensures \result!=\null;
  @   ensures isShallowCopyOfAny{Post,Pre}(*\result,*arg);
  @ behavior no_allocation:
  @   assumes !is_allocable{Pre}((size_t)sizeof(Any),(size_t)1);
  @   assigns \nothing;
  @   ensures \result==\null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Any* copyAny(const Any* arg); /* shallow copy */
Any* copyAnyDeep(const Any*); /* deep copy */


/*@ requires size+1 <= UINT32_MAX;
  @ behavior allocation:
  @   assumes is_allocable{Pre}((size_t)(size+1),(size_t)1);
  @   assigns __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   ensures allocation_footprint{Pre,Post}((size_t)(size+1),(size_t)1);
  @   ensures is_allocated_size{Post}((char*)\result,(size_t)(size+1));
  @   ensures \forall char* p; (is_allocated{Pre}(p) ==> p != (char*)\result);
  @   ensures \forall char* p, size_t n; is_allocated_size{Pre}(p,n) ==> is_allocated_size{Post}(p,n);
  @   ensures \forall size_t i; 0 <= i <= \at(__num_allocated,Post)-1 ==> \at(__allocated[i],Post) == \at(__allocated[i],Pre);
  @   ensures __allocated[__num_allocated-1] == (char*)\result;
  @   ensures \valid((char*)\result+(0..size));
  @   ensures \separated(__allocated+(0..MAX_ALLOCS-1),(char*)\result+(0..size));
  @   ensures \forall char** a, size_t s; is_allocated_size{Pre}((char*)a, s) ==> 
  @                     \separated((char*)\result+(0..(size_t)(size)),a+(0..s-1));
  @   ensures \result!=\null;
  @ behavior no_allocation:
  @   assumes !is_allocable{Pre}((size_t)(size+1),(size_t)1);
  @   assigns \nothing;
  @   ensures \result==\null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @   ensures \forall char* p, size_t n; is_allocated_size{Pre}(p,n) ==> is_allocated_size{Post}(p,n);
  @   ensures \forall size_t i; 0 <= i <= \at(__num_allocated,Post) ==> \at(__allocated[i],Post) == \at(__allocated[i],Pre);
  @   ensures !is_allocable{Pre}((size_t)(size+1),(size_t)1);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Name allocName(size_t size);

/*@
  @   requires  is_allocated{Pre}((char*)ptr);
  @   assigns  __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   frees ptr;
  @   ensures !is_allocated{Post}((char*)ptr);
  @   ensures old_allocated{Pre,Post}((char *)ptr);
  @   ensures \forall size_t s; is_allocated_size{Pre}((char*)ptr,s) ==> \at(__size_allocated,Post) == \at(__size_allocated,Pre) - s;
  @   ensures allocation_footprint{Post,Pre}((size_t)(\block_length{Pre}((char*)ptr)), (size_t)1);
  @*/
void freeName(Name ptr); 
void freeNameZ(Name*);

/*@ requires valid_string(arg);
  @ requires strlen(arg)+1 <= UINT32_MAX;
  @ behavior allocation:
  @   assumes is_allocable{Pre}((size_t)(strlen(arg)+1),(size_t)1);
  @   assigns __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   ensures allocation_footprint{Pre,Post}((size_t)(strlen(arg)+1),(size_t)1);
  @   ensures is_allocated_size{Post}((char*)\result,(size_t)strlen(arg));
  @   ensures \valid((char*)\result+(0..strlen(arg)));
  @   ensures \result!=\null;
  @ behavior no_allocation:
  @   assumes !is_allocable{Pre}((size_t)(strlen(arg)+1),(size_t)1);
  @   assigns \nothing;
  @   ensures \result==\null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Name makeName(const Name arg); /* shallow copy = deep copy*/


/*@ requires valid_string(arg);
  @ requires strlen(arg)+1 <= UINT32_MAX;
  @ behavior allocation:
  @   assumes is_allocable{Pre}((size_t)(strlen(arg)+1),(size_t)1);
  @   assigns __allocated[__num_allocated], __size_allocated, __num_allocated;
  @   ensures allocation_footprint{Pre,Post}((size_t)(strlen(arg)+1),(size_t)1);
  @   ensures is_allocated_size{Post}((char*)\result,(size_t)strlen(arg));
  @   ensures \valid((char*)\result+(0..strlen(arg)));
  @   ensures \result!=\null;
  @ behavior no_allocation:
  @   assumes !is_allocable{Pre}((size_t)(strlen(arg)+1),(size_t)1);
  @   assigns \nothing;
  @   ensures \result==\null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Name copyName(const Name arg); /* shallow copy = deep copy*/
Name copyNameDeep(const Name); /* shallow copy = deep copy*/

void die(const char *reason);
void freeBindParts(Bind b);

#endif
