/****************************************************************************
 * Copyright (C) 2016 by University of Warsaw                               *
 *                                                                          *
 * This file is part of IntuitionTC.                                        *
 *                                                                          *
 ****************************************************************************/
/**
 * @file checker.h
 * @brief Definitions of the base structures that represent the formulas and
 *        proofs.
 * @author Aleksy Schubert
 * @author Maciej Zielenkiewicz
 * @author Marcin Benke
 */
#ifndef _CHECKER_H
#define _CHECKER_H  1
#include <stdlib.h>
#ifdef _FRAMAC_
#include <__fc_string_axiomatic.h>
#include "errorMessages.h"
#endif

extern const size_t DEFAULT_SIZE;
//@ predicate default_size_inv(size_t ds) = ds == 100;


/*@ axiomatic string_additional {
  @   lemma strcmp_equal{L}:
  @       \forall char *s; valid_string(s) ==> strcmp(s, s) == 0;
  @ }
  @*/

/** Predefined type names to enable cyclic type definitions below. */
struct Type;
typedef struct Type Type;
struct Kind;
typedef struct Kind Kind;
struct Term;
typedef struct Term Term;
struct Bind;
typedef struct Bind Bind;
struct Any;
typedef struct Any Any;

/** @brief The type to structurally represent identifier names. */
typedef char* Name;
/*@ predicate eqName(Name n1, Name n2) = 
  @   (n1 == n2) || (n1 != \null && n2 != \null && strcmp(n1,n2) == 0);
  @*/

/** @brief Values to indicate different species of #Kind elements. */
/* data Kind = Kstar 
           | Kpi Name Type Kind */
typedef enum { eKstar, eKpi } eKind;

/** @brief The representation of the star #Kind. */
typedef struct Kstar { int nothing; } Kstar;
// something must be inside a structure for WHY3 to digest

/** @brief The representation of the pi #Kind. */
typedef struct Kpi {
    Name name;
    Type* type;
    Kind* kind;
} Kpi;

/** @brief The representation of the kinds in the typesystem. */
struct Kind {
    eKind  which;
    Kstar kStar;
    Kpi   kPi;
};

/** @brief Values to indicate different species of #Type elements. */
/*            
data Type = Tvar Name 
          | Tall Name Type Type 
          | Tapp Type Term
          | Texi Name Type Type
          | Tand Type Type
          | Tor Type Type  
          | Tbot 
*/
typedef enum { etVar, etAll, etApp, etExi, etAnd, etOr, etBot } eType;

/** @brief The representation of the type variable in #Type. */
/* data Type = Tvar Name */
typedef struct Tvar {
    Name name;
} Tvar;


/** @brief The representation of the universal quantifier #Type. */
/*           | Tall Name Type Type */
typedef struct Tall {
    Name  name;
    Type* type1;
    Type* type2;
} Tall;

/** @brief The representation of the application of #Type to #Term;
 *         it is one of the species of #Type. */
/*          | Tapp Type Term */
typedef struct Tapp {
    Type* type;
    Term* term;
} Tapp;

/** @brief The representation of the existential quantifier #Type. */
/*          | Texi Name Type Type */
typedef struct Texi {
    Name  name;
    Type* type1;
    Type* type2;
} Texi;

/** @brief The representation of the conjunction #Type. */
/*          | Tand Type Type */
typedef struct Tand {
    Type* type1;
    Type* type2;
} Tand;

/** @brief The representation of the disjunction #Type. */
/*          | Tor Type Type */
typedef struct Tor {
    Type* type1;
    Type* type2;
} Tor;

/** @brief The representation of the bottom/falsity #Type. */
/*          | Tbot */
typedef struct Tbot { int nothing; } Tbot;
// something must be inside a structure for WHY3 to digest

/** @brief The representation of the types in the typesystem. */
/*            
data Type = Tvar Name 
          | Tall Name Type Type 
          | Tapp Type Term
          | Texi Name Type Type
          | Tand Type Type
          | Tor Type Type  
          | Tbot 
*/
struct Type {
    eType  which;
    Tvar   tVar;
    Tall   tAll;
    Tapp   tApp;
    Texi   tExi;
    Tand   tAnd;
    Tor    tOr;
    Tbot   tBot;
};


/** @brief Values to indicate different species of #Term elements. */
/*
data Term = Mvar Name
          | Mapp Term Term 
          | Mlam Name Type Term
          | Mwit Type Term Term -- [m1,m2]_{∃ x:phi1.phi2}
          | Mabs Name Type Name Type Term Term
            -- abstract <x:phi1,y:phi2> = m1 in m2
          | Mtup Type Term Term
          | Mpi1 Term
          | Mpi2 Term
          | Min1 Type Term
          | Min2 Type Term
          | Mcas Term (Name,Type,Term) (Name,Type,Term)
*/
typedef enum {emVar, emApp, emLam, emWit, emAbs, emTup, emPi1, emPi2, emIn1, 
              emIn2, emCas, emEps} eTerm;

/** @brief The representation of the proof term variable/axiom in #Term. */
/* data Term = Mvar Name */
typedef struct Mvar {
	Name	name;
} Mvar;

/** @brief The representation of the application of a #Term to a #Term;
 *         it is one of the species of #Term. */
/*          | Mapp Term Term */
typedef struct Mapp {
	Term*	term1;
	Term*	term2;
} Mapp;

/** @brief The representation of the lambda abstraction #Term. */
/*          | Mlam Name Type Term*/
typedef	struct Mlam {
	Name	name;
	Type*	type;
	Term*	term;
} Mlam;

/** @brief The representation of the exists
 *   (i.e. existential quantifier introduction) #Term. */
/*          | Mwit Type Term Term -- [m1,m2]_{∃ x:phi1.phi2}*/
typedef	struct Mwit {
	Type*	type;
	Term*	term1;
	Term*	term2;
} Mwit;

/** @brief The representation of the abstraction
 *   (i.e. existential quantifier elimination) #Term. */
/*          | Mabs Name Type Name Type Term Term
              -- abstract <x:phi1,y:phi2> = m1 in m2*/
typedef	struct Mabs {
	Name	name1;
	Type*	type2;
	Name	name3;
	Type*	type4;
	Term*	term5;
	Term*	term6;
} Mabs;

/** @brief The representation of the tuple
 *   (i.e. conjunction introduction) #Term. */
/*          | Mtup Type Term Term*/
typedef	struct Mtup {
	Type*	type;
	Term*	term1;
	Term*	term2;
} Mtup;

/** @brief The representation of the first projection
 *   (i.e. conjunction elimination) #Term. */
/*          | Mpi1 Term*/
typedef struct Mpi1 {
	Term*	term;
} Mpi1;

/** @brief The representation of the second projection
 *   (i.e. conjunction elimination) #Term. */
/*          | Mpi2 Term*/
typedef struct Mpi2 {
	Term*	term;
} Mpi2;

/** @brief The representation of the first injection
 *   (i.e. disjunction introduction) #Term. */
/*          | Min1 Type Term*/
typedef struct Min1 {
	Type*	type;
	Term*	term;
} Min1;

/** @brief The representation of the second injection
 *   (i.e. disjunction introduction) #Term. */
/*          | Min2 Type Term*/
typedef struct Min2 {
	Type*	type;
	Term*	term;
} Min2;

/** @brief The representation of the case subexpression. */
/*          | Mcas Term (Name,Type,Term) (Name,Type,Term)*/
typedef struct McasIn {
	Name	name;
	Type*	type;
	Term*	term;
} McasIn;

/** @brief The representation of the case analysis
 *   (i.e. disjunction elimination) #Term. */
typedef struct Mcas {
	Term*	term;
	McasIn	casIn1;
	McasIn	casIn2;
} Mcas;

/** @brief The representation of the ex falso quod libet
 *   (i.e. falsity elimination) #Term. */
/*          | Meps Type Term*/
typedef struct Meps {
	Type*	type;
	Term*	term;
} Meps;

/** @brief The representation of the proof terms in the typesystem. */
/* 
data Term = Mvar Name
          | Mapp Term Term 
          | Mlam Name Type Term
          | Mwit Type Term Term -- [m1,m2]_{∃ x:phi1.phi2}
          | Mabs Name Type Name Type Term Term
            -- abstract <x:phi1,y:phi2> = m1 in m2
          | Mtup Type Term Term
          | Mpi1 Term
          | Mpi2 Term
          | Min1 Type Term
          | Min2 Type Term
          | Mcas Term (Name,Type,Term) (Name,Type,Term)
*/
struct Term {
	eTerm	which;
	Mvar	mVar;
	Mapp	mApp;
	Mlam	mLam;
	Mwit	mWit;
	Mabs	mAbs;
	Mtup	mTup;
	Mpi1	mPi1;
	Mpi2	mPi2;
	Min1	mIn1;
	Min2	mIn2;
	Mcas	mCas;
	Meps	mEps;
};

/** @brief The structure to represent binders. (For future use.) */
struct Bind {
	Name name;
	Type*type;
	Any *any;
};


/** @brief Values to indicate different species of data held in 
 *  #Any structures. */
typedef enum { eAnyType, eAnyTerm, eAnyKind, eAnyName, eAnyBind} eAny;

/** @brief The wrapper that makes it possible to store in #envmap
 *  different species of elements that occur in typechecking
 *  process. */
struct Any {
	eAny	which;
	Type*	mType;
        Term*	mTerm;
	Kind*	mKind;
	Name	mName;
	Bind	mBind; 
};




/*@ axiomatic checker { 
  @
  @ predicate valid_Type_ptr{L}(Type *t);
  @
  @ predicate valid_Kind_ptr{L}(Kind *k);
  @
  @ predicate valid_Term_ptr{L}(Term *t);
  @
  @ predicate valid_Name{L}(Name n) = valid_string{L}(n);
  @
  @ predicate valid_Kstar{L}(Kstar k) = \true;
  @
  @ predicate valid_Kpi(Kpi k) =
  @    valid_Name(k.name) &&
  @    valid_Type_ptr(k.type) &&
  @    valid_Kind_ptr(k.kind);
  @
  @ predicate valid_Kind{L}(Kind k) = 
  @    (k.which == eKstar || k.which == eKpi) &&
  @    (k.which == eKstar ==> valid_Kstar(k.kStar)) &&
  @    (k.which == eKpi ==> valid_Kpi(k.kPi));
  @
  @ predicate isShallowCopyOfKind{L,M}(Kind k1, Kind k2) = 
  @    \at(k1.which,L) == \at(k2.which,M) &&
  @    \at(k1.kStar,L) == \at(k2.kStar,M) &&
  @    \at(k1.kPi,L) == \at(k2.kPi,M);
  @
  @ axiom valid_Kind_ptr_def{L}: \forall Kind *k; 
  @         valid_Kind_ptr{L}(k) <==> \valid(k) && valid_Kind{L}(*k);
  @
  @
  @ predicate valid_Tvar{L}(Tvar t) =
  @     valid_Name(t.name);
  @
  @
  @ predicate valid_Tall{L}(Tall t) = 
  @     valid_Name(t.name) &&
  @     \valid(t.type1) &&
  @     valid_Type_ptr(t.type1) &&
  @     \valid(t.type2) &&
  @     valid_Type_ptr(t.type2);
  @
  @
  @ predicate valid_Tapp{L}(Tapp t) = 
  @     \valid(t.type) &&
  @     valid_Type_ptr(t.type) &&
  @     \valid(t.term) &&
  @     valid_Term_ptr(t.term);
  @
  @
  @ predicate valid_Texi{L}(Texi t) = 
  @     valid_Name(t.name) &&
  @     \valid(t.type1) &&
  @     valid_Type_ptr(t.type1) &&
  @     \valid(t.type2) &&
  @     valid_Type_ptr(t.type2);
  @
  @
  @ predicate valid_Tand{L}(Tand t) = 
  @     \valid(t.type1) &&
  @     valid_Type_ptr(t.type1) &&
  @     \valid(t.type2) &&
  @     valid_Type_ptr(t.type2);
  @
  @
  @ predicate valid_Tor{L}(Tor t) = 
  @     \valid(t.type1) &&
  @     valid_Type_ptr(t.type1) &&
  @     \valid(t.type2) &&
  @     valid_Type_ptr(t.type2);
  @
  @
  @ predicate valid_Tbot{L}(Tbot t) = \true;
  @
  @
  @ predicate valid_Type{L}(struct Type t) = 
  @    (t.which == etVar || t.which == etAll || t.which == etApp ||
  @     t.which == etExi || t.which == etAnd || t.which == etOr ||
  @     t.which == etBot) &&
  @    (t.which == etVar ==> valid_Tvar(t.tVar)) &&
  @    (t.which == etAll ==> valid_Tall(t.tAll)) &&
  @    (t.which == etApp ==> valid_Tapp(t.tApp)) &&
  @    (t.which == etExi ==> valid_Texi(t.tExi)) &&
  @    (t.which == etAnd ==> valid_Tand(t.tAnd)) &&
  @    (t.which == etOr ==> valid_Tor(t.tOr)) &&
  @    (t.which == etBot ==> valid_Tbot(t.tBot));
  @
  @ predicate isShallowCopyOfType{L,M}(Type t1, Type t2) = 
  @    \at(t1.which,L) == \at(t2.which,M) &&
  @    \at(t1.tVar,L)  == \at(t2.tVar,M) &&
  @    \at(t1.tAll,L)  == \at(t2.tAll,M) &&
  @    \at(t1.tApp,L)  == \at(t2.tApp,M) &&
  @    \at(t1.tExi,L)  == \at(t2.tExi,M) &&
  @    \at(t1.tAnd,L)  == \at(t2.tAnd,M) &&
  @    \at(t1.tOr,L)   == \at(t2.tOr,M) &&
  @    \at(t1.tBot,L)  == \at(t2.tBot,M);
  @
  @ axiom valid_Type_ptr_def{L}: \forall Type *t; 
  @         valid_Type_ptr{L}(t) <==> \valid(t) && valid_Type{L}(*t);
  @
  @ predicate valid_Mvar{L}(Mvar m) =
  @     valid_Name(m.name);
  @
  @ predicate valid_Mapp{L}(Mapp m) =
  @     \valid(m.term1) &&
  @     valid_Term_ptr(m.term1) &&
  @     \valid(m.term2) &&
  @     valid_Term_ptr(m.term2);
  @
  @ predicate valid_Mlam{L}(Mlam m) = 
  @     valid_Name(m.name) &&
  @     \valid(m.type) &&
  @     valid_Type_ptr(m.type) &&
  @     \valid(m.term) &&
  @     valid_Term_ptr(m.term);
  @
  @ predicate valid_Mwit{L}(Mwit m) = 
  @     \valid(m.type) &&
  @     valid_Type_ptr(m.type) &&
  @     \valid(m.term1) &&
  @     valid_Term_ptr(m.term1) &&
  @     \valid(m.term2) &&
  @     valid_Term_ptr(m.term2);
  @ 
  @ predicate valid_Mabs{L}(Mabs m) = 
  @     valid_Name(m.name1) &&
  @     \valid(m.type2) &&
  @     valid_Type_ptr(m.type2) &&
  @     valid_Name(m.name3) &&
  @     \valid(m.type4) &&
  @     valid_Type_ptr(m.type4) &&
  @     \valid(m.term5) &&
  @     valid_Term_ptr(m.term5) &&
  @     \valid(m.term6) &&
  @     valid_Term_ptr(m.term6);
  @
  @ predicate valid_Mtup{L}(Mtup m) = 
  @     \valid(m.type) &&
  @     valid_Type_ptr(m.type) &&
  @     \valid(m.term1) &&
  @     valid_Term_ptr(m.term1) &&
  @     \valid(m.term2) &&
  @     valid_Term_ptr(m.term2);
  @
  @ predicate valid_Mpi1{L}(Mpi1 m) = 
  @     \valid(m.term) &&
  @     valid_Term_ptr(m.term);
  @
  @ predicate valid_Mpi2{L}(Mpi2 m) = 
  @     \valid(m.term) &&
  @     valid_Term_ptr(m.term);
  @
  @ predicate valid_Min1{L}(Min1 m) = 
  @     \valid(m.type) &&
  @     valid_Type_ptr(m.type) &&
  @     \valid(m.term) &&
  @     valid_Term_ptr(m.term);
  @
  @ predicate valid_Min2{L}(Min2 m) = 
  @     \valid(m.type) &&
  @     valid_Type_ptr(m.type) &&
  @     \valid(m.term) &&
  @     valid_Term_ptr(m.term);
  @
  @ predicate valid_McasIn{L}(McasIn m) = 
  @     valid_Name(m.name) &&
  @     \valid(m.type) &&
  @     valid_Type_ptr(m.type) &&
  @     \valid(m.term) &&
  @     valid_Term_ptr(m.term);
  @
  @ predicate valid_Mcas{L}(Mcas m) = 
  @     \valid(m.term) &&
  @     valid_Term_ptr(m.term) &&
  @     valid_McasIn(m.casIn1) &&
  @     valid_McasIn(m.casIn2);
  @
  @ predicate valid_Meps{L}(Meps m) = 
  @     \valid(m.type) &&
  @     valid_Type_ptr(m.type) &&
  @     \valid(m.term) &&
  @     valid_Term_ptr(m.term);
  @
  @ predicate valid_Term{L}(struct Term t) = 
  @    (t.which == emVar || t.which == emApp || t.which == emLam ||
  @     t.which == emWit || t.which == emAbs || t.which == emTup ||
  @     t.which == emPi1 || t.which == emPi2 || t.which == emIn1 || 
  @     t.which == emIn2 || t.which == emCas || t.which == emEps) &&
  @    (t.which == emVar ==> valid_Mvar(t.mVar)) && 
  @    (t.which == emApp ==> valid_Mapp(t.mApp)) && 
  @    (t.which == emLam ==> valid_Mlam(t.mLam)) && 
  @    (t.which == emWit ==> valid_Mwit(t.mWit)) && 
  @    (t.which == emAbs ==> valid_Mabs(t.mAbs)) &&
  @    (t.which == emTup ==> valid_Mtup(t.mTup)) && 
  @    (t.which == emPi1 ==> valid_Mpi1(t.mPi1)) && 
  @    (t.which == emPi2 ==> valid_Mpi2(t.mPi2)) && 
  @    (t.which == emIn1 ==> valid_Min1(t.mIn1)) && 
  @    (t.which == emIn2 ==> valid_Min2(t.mIn2)) && 
  @    (t.which == emCas ==> valid_Mcas(t.mCas)) && 
  @    (t.which == emEps ==> valid_Meps(t.mEps));
  @
  @ predicate isShallowCopyOfTerm{L,M}(Term t1, Term t2) = 
  @    \at(t1.which,L) == \at(t2.which,M) &&
  @    \at(t1.mVar,L) == \at(t2.mVar,M) &&
  @    \at(t1.mApp,L) == \at(t2.mApp,M) &&
  @    \at(t1.mLam,L) == \at(t2.mLam,M) &&
  @    \at(t1.mWit,L) == \at(t2.mWit,M) &&
  @    \at(t1.mAbs,L) == \at(t2.mAbs,M) &&
  @    \at(t1.mTup,L) == \at(t2.mTup,M) &&
  @    \at(t1.mPi1,L) == \at(t2.mPi1,M) &&
  @    \at(t1.mPi2,L) == \at(t2.mPi2,M) &&
  @    \at(t1.mIn1,L) == \at(t2.mIn1,M) &&
  @    \at(t1.mIn2,L) == \at(t2.mIn2,M) &&
  @    \at(t1.mCas,L) == \at(t2.mCas,M) &&
  @    \at(t1.mEps,L) == \at(t2.mEps,M);
  @
  @ axiom valid_Term_ptr_def: 
  @  \forall Term *t; valid_Term_ptr(t) <==>
  @    \valid(t) && valid_Term(*t);
  @
  @
  @ predicate valid_Any{L}(struct Any t) = 
  @    (t.which == eAnyType || t.which == eAnyTerm || t.which == eAnyKind ||
  @     t.which == eAnyName || t.which == eAnyBind) &&
  @    (t.which == eAnyType ==> valid_Type_ptr(t.mType)) &&
  @    (t.which == eAnyTerm ==> valid_Term_ptr(t.mTerm)) &&
  @    (t.which == eAnyKind ==> valid_Kind_ptr(t.mKind)) &&
  @    (t.which == eAnyName ==> valid_Name(t.mName)); // &&
  @    //(t.which == eAnyBind ==> valid_Bind(t.mBind));
  @
  @ predicate isShallowCopyOfAny{L,M}(Any a1, Any a2) = 
  @    \at(a1.which,L) == \at(a2.which,M) &&
  @    \at(a1.mType,L) == \at(a2.mType,M) &&
  @    \at(a1.mTerm,L) == \at(a2.mTerm,M) &&
  @    \at(a1.mKind,L) == \at(a2.mKind,M) &&
  @    \at(a1.mName,L) == \at(a2.mName,M) &&
  @    \at(a1.mBind,L) == \at(a2.mBind,M);
  @
  @ predicate valid_Any_ptr{L}(Any *t) =
  @    t != \null &&
  @    \valid(t) &&
  @    valid_Any{L}(*t);
  @
  @ predicate valid_Bind{L}(Bind b) = 
  @    (valid_Name{L}(b.name)) &&
  @    (valid_Type_ptr{L}(b.type)) &&
  @    (valid_Any_ptr{L}(b.any));

  } //axiomatic checker @*/

//----------------------------------------------------------------------

/*@ axiomatic wellfounded {
  @ 
  @ 
  @ logic integer depth_Type(Type *t);
  @ logic integer depth_Term(Term *t);
  @ logic integer depth_Kind(Kind *t);
  @ logic integer depth_Any(Any *a);
  @
  @  // for Type
  @ 
  @ logic integer depth_Tvar(Tvar t) = 1;
  @
  @ logic integer depth_Tall(Tall t) = 1 +
  @     maximum(depth_Type(t.type1), depth_Type(t.type2));
  @
  @ logic integer depth_Tapp(Tapp t) = 1 + 
  @     maximum(depth_Type(t.type), depth_Term(t.term));
  @
  @ logic integer depth_Texi(Texi t) = 1 +
  @     maximum(depth_Type(t.type1), depth_Type(t.type2));
  @
  @ logic integer depth_Tand(Tand t) = 1 +
  @     maximum(depth_Type(t.type1), depth_Type(t.type2));
  @
  @ logic integer depth_Tor(Tor t) = 1 +
  @     maximum(depth_Type(t.type1), depth_Type(t.type2));
  @
  @ logic integer depth_Tbot(Tbot t) = 1;
  @ 
  @  // for Term
  @ 
  @ logic integer depth_Mvar(Mvar t) = 1;
  @
  @ logic integer depth_Mapp(Mapp t) = 1 +
  @     maximum(depth_Term(t.term1), depth_Term(t.term2));
  @
  @ logic integer depth_Mlam(Mlam t) = 1 +
  @     maximum(depth_Type(t.type), depth_Term(t.term));
  @
  @ logic integer depth_Mwit(Mwit t) = 1 +
  @     maximum(depth_Type(t.type),
  @       maximum(depth_Term(t.term1), depth_Term(t.term2)));
  @
  @ logic integer depth_Mabs(Mabs t) = 1 +
  @     maximum(
  @       maximum(depth_Type(t.type2), depth_Type(t.type4)),
  @       maximum(depth_Term(t.term5), depth_Term(t.term6)));
  @
  @ logic integer depth_Mtup(Mtup t) = 1 +
  @     maximum(depth_Term(t.term1), depth_Term(t.term2));
  @
  @ logic integer depth_Mpi1(Mpi1 t) = 1 +
  @     depth_Term(t.term);
  @
  @ logic integer depth_Mpi2(Mpi2 t) = 1 +
  @     depth_Term(t.term);
  @
  @ logic integer depth_Min1(Min1 t) = 1 +
  @     maximum(depth_Type(t.type), depth_Term(t.term));
  @
  @ logic integer depth_Min2(Min2 t) = 1 +
  @     maximum(depth_Type(t.type), depth_Term(t.term));
  @
  @ logic integer depth_McasIn(McasIn t) = 
  @     maximum(depth_Type(t.type), depth_Term(t.term));
  @
  @ logic integer depth_Mcas(Mcas t) = 
  @     maximum(depth_Term(t.term),
  @       maximum(depth_McasIn(t.casIn1), 
  @               depth_McasIn(t.casIn2)));
  @
  @ logic integer depth_Meps(Meps t) = 1 +
  @     maximum(depth_Type(t.type), depth_Term(t.term));
  @
  @ logic integer depth_Bind(Bind b) = 1 +
  @     maximum(depth_Type(b.type), depth_Any(b.any));
  @
  @ 
  @ axiom depth_Type_def{M}: 
  @    \forall Type *t;
  @      ((t->which==etVar) ==> depth_Type(t) == depth_Tvar(t->tVar)) &&
  @      ((t->which==etAll) ==> depth_Type(t) == depth_Tall(t->tAll)) &&
  @      ((t->which==etApp) ==> depth_Type(t) == depth_Tapp(t->tApp)) &&
  @      ((t->which==etExi) ==> depth_Type(t) == depth_Texi(t->tExi)) &&
  @      ((t->which==etAnd) ==> depth_Type(t) == depth_Tand(t->tAnd)) &&
  @      ((t->which==etOr ) ==> depth_Type(t) == depth_Tor(t->tOr))   &&
  @      ((t->which==etBot) ==> depth_Type(t) == depth_Tbot(t->tBot)) &&
  @      ((\exists integer n; depth_Type(t)==n) ==> 
  @          (t->which==etVar) ||
  @          (t->which==etAll) ||
  @          (t->which==etApp) ||
  @          (t->which==etExi) ||
  @          (t->which==etAnd) ||
  @          (t->which==etOr ) ||
  @          (t->which==etBot));
  @
  @
  @ axiom depth_Term_def{M}:
  @    \forall Term *t;
  @      ((t->which==emVar) ==> depth_Term(t) == depth_Mvar(t->mVar)) &&
  @      ((t->which==emApp) ==> depth_Term(t) == depth_Mapp(t->mApp)) &&
  @      ((t->which==emLam) ==> depth_Term(t) == depth_Mlam(t->mLam)) &&
  @      ((t->which==emWit) ==> depth_Term(t) == depth_Mwit(t->mWit)) &&
  @      ((t->which==emAbs) ==> depth_Term(t) == depth_Mabs(t->mAbs)) &&
  @      ((t->which==emTup) ==> depth_Term(t) == depth_Mtup(t->mTup)) &&
  @      ((t->which==emPi1) ==> depth_Term(t) == depth_Mpi1(t->mPi1)) &&
  @      ((t->which==emPi2) ==> depth_Term(t) == depth_Mpi2(t->mPi2)) &&
  @      ((t->which==emIn1) ==> depth_Term(t) == depth_Min1(t->mIn1)) &&
  @      ((t->which==emIn2) ==> depth_Term(t) == depth_Min2(t->mIn2)) &&
  @      ((t->which==emCas) ==> depth_Term(t) == depth_Mcas(t->mCas)) &&
  @      ((t->which==emEps) ==> depth_Term(t) == depth_Meps(t->mEps)) &&
  @      ((\exists integer n; depth_Term(t)==n) ==> 
  @         ((t->which==emVar) ||
  @          (t->which==emApp) ||
  @          (t->which==emLam) ||
  @          (t->which==emWit) ||
  @          (t->which==emAbs) ||
  @          (t->which==emTup) ||
  @          (t->which==emPi1) ||
  @          (t->which==emPi2) ||
  @          (t->which==emIn1) ||
  @          (t->which==emIn2) ||
  @          (t->which==emCas) ||
  @          (t->which==emEps)));
  @
  @
  @ axiom depth_Kind_def{M}:
  @    \forall Kind *t;
  @      ((t->which==eKstar) ==> depth_Kind(t) == 1) &&
  @      ((t->which==eKpi) ==> depth_Kind(t) == 
  @                 1 + maximum(depth_Type(t->kPi.type),
  @                             depth_Kind(t->kPi.kind))) &&
  @      ((\exists integer n; depth_Kind(t)==n) ==> 
  @         (t->which==eKstar || (t->which==eKpi)));
  @
  @
  @ axiom depth_Any_def{M}:
  @    \forall Any* a;
  @        ((a->which==eAnyType) ==> depth_Any(a) == depth_Type(a->mType)) &&
  @        ((a->which==eAnyTerm) ==> depth_Any(a) == depth_Term(a->mTerm)) &&
  @        ((a->which==eAnyKind) ==> depth_Any(a) == depth_Kind(a->mKind)) &&
  @        ((a->which==eAnyName) ==> depth_Any(a) == 0) &&
  @        ((a->which==eAnyBind) ==> depth_Any(a) == depth_Bind(a->mBind)) &&
  @        ((\exists integer n; depth_Any(a) ==n) ==> 
  @                                 (a->which==eAnyType ||
  @                                  a->which==eAnyTerm ||
  @                                  a->which==eAnyKind ||
  @                                  a->which==eAnyName ||
  @                                  a->which==eAnyBind));
  @ 
  @ } 
  @*/

//----------------------------------------------------------------------

/*@ axiomatic checker_separation { 
  @
  @ predicate separated_Type{L}(char** keys, size_t len, Type *t);
  @ predicate separated_Term{L}(char** keys, size_t len, Term *t);
  @
  @ predicate separated_Name{L}(char** keys, size_t len, Name n) =
  @     \separated(keys+(0..(sizeof(char*)*len)-1), n+(0..strlen(n)));
  @
  @ predicate separated_Tvar{L}(char** keys, size_t len, Tvar t) =
  @     \separated(keys+(0..(sizeof(char*)*len)-1), t.name+(0..strlen(t.name)));
  @
  @ predicate separated_Tall{L}(char** keys, size_t len, Tall t) = 
  @     separated_Name(keys, len, t.name) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), t.type1) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), t.type2) &&
  @     separated_Type(keys, len, t.type1) &&
  @     separated_Type(keys, len, t.type2);
  @
  @
  @ predicate separated_Tapp{L}(char** keys, size_t len, Tapp t) = 
  @     separated_Type(keys, len, t.type) &&
  @     separated_Term(keys, len, t.term);
  @
  @
  @ predicate separated_Texi{L}(char** keys, size_t len, Texi t) = 
  @     separated_Name(keys, len, t.name) &&
  @     separated_Type(keys, len, t.type1) &&
  @     separated_Type(keys, len, t.type2);
  @
  @ predicate separated_Tand{L}(char** keys, size_t len, Tand t) = 
  @     separated_Type(keys, len, t.type1) &&
  @     separated_Type(keys, len, t.type2);
  @
  @ predicate separated_Tor{L}(char** keys, size_t len, Tor t) = 
  @     separated_Type(keys, len, t.type1) &&
  @     separated_Type(keys, len, t.type2);
  @
  @ predicate separated_Tbot{L}(char** keys, size_t len, Tbot t) = \true;
  @
  @  axiom separated_Type_def{L}:
  @    \forall char** keys, size_t len, Type *t;
  @      separated_Type(keys, len, t) <==>
  @      \separated(keys+(0..(sizeof(char*)*len)-1), t) &&
  @      (t->which == etVar ==> separated_Tvar(keys, len, t->tVar)) &&
  @      (t->which == etAll ==> separated_Tall(keys, len, t->tAll)) &&
  @      (t->which == etApp ==> separated_Tapp(keys, len, t->tApp)) &&
  @      (t->which == etExi ==> separated_Texi(keys, len, t->tExi)) &&
  @      (t->which == etAnd ==> separated_Tand(keys, len, t->tAnd)) &&
  @      (t->which == etOr ==>  separated_Tor(keys, len, t->tOr)) &&
  @      (t->which == etBot ==> separated_Tbot(keys, len, t->tBot));
  @
  @
  @ predicate separated_Mvar{L}(char** keys, size_t len, Mvar m) =
  @     separated_Name(keys, len, m.name);
  @
  @ predicate separated_Mapp{L}(char** keys, size_t len, Mapp m) =
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.term1) &&
  @     separated_Term(keys, len, m.term1) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.term2) &&
  @     separated_Term(keys, len, m.term2);
  @
  @ predicate separated_Mlam{L}(char** keys, size_t len, Mlam m) = 
  @     separated_Name(keys, len, m.name) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.type) &&
  @     separated_Type(keys, len, m.type) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.term) &&
  @     separated_Term(keys, len, m.term);
  @
  @ predicate separated_Mwit{L}(char** keys, size_t len, Mwit m) = 
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.type) &&
  @     separated_Type(keys, len, m.type) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.term1) &&
  @     separated_Term(keys, len, m.term1) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.term2) &&
  @     separated_Term(keys, len, m.term2);
  @ 
  @ predicate separated_Mabs{L}(char** keys, size_t len, Mabs m) = 
  @     separated_Name(keys, len, m.name1) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.type2) &&
  @     separated_Type(keys, len, m.type2) &&
  @     separated_Name(keys, len, m.name3) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.type4) &&
  @     separated_Type(keys, len, m.type4) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.term5) &&
  @     separated_Term(keys, len, m.term5) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.term6) &&
  @     separated_Term(keys, len, m.term6);
  @
  @ predicate separated_Mtup{L}(char** keys, size_t len, Mtup m) = 
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.type) &&
  @     separated_Type(keys, len, m.type) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.term1) &&
  @     separated_Term(keys, len, m.term1) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.term2) &&
  @     separated_Term(keys, len, m.term2);
  @
  @ predicate separated_Mpi1{L}(char** keys, size_t len, Mpi1 m) = 
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.term) &&
  @     separated_Term(keys, len, m.term);
  @
  @ predicate separated_Mpi2{L}(char** keys, size_t len, Mpi2 m) = 
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.term) &&
  @     separated_Term(keys, len, m.term);
  @
  @ predicate separated_Min1{L}(char** keys, size_t len, Min1 m) = 
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.type) &&
  @     separated_Type(keys, len, m.type) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.term) &&
  @     separated_Term(keys, len, m.term);
  @
  @ predicate separated_Min2{L}(char** keys, size_t len, Min2 m) = 
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.type) &&
  @     separated_Type(keys, len, m.type) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.term) &&
  @     separated_Term(keys, len, m.term);
  @
  @ predicate separated_McasIn{L}(char** keys, size_t len, McasIn m) = 
  @     separated_Name(keys, len, m.name) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.type) &&
  @     separated_Type(keys, len, m.type) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.term) &&
  @     separated_Term(keys, len, m.term);
  @
  @ predicate separated_Mcas{L}(char** keys, size_t len, Mcas m) = 
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.term) &&
  @     separated_Term(keys, len, m.term) &&
  @     separated_McasIn(keys, len, m.casIn1) &&
  @     separated_McasIn(keys, len, m.casIn2);
  @
  @ predicate separated_Meps{L}(char** keys, size_t len, Meps m) = 
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.type) &&
  @     separated_Type(keys, len, m.type) &&
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.term) &&
  @     separated_Term(keys, len, m.term);
  @
  @  axiom separated_Term_def{L}:
  @    \forall char** keys, size_t len, Term *t;
  @      separated_Term(keys, len, t) <==>
  @      \separated(keys+(0..(sizeof(char*)*len)-1), t) &&
  @      (t->which == emVar ==> separated_Mvar(keys, len, t->mVar)) && 
  @      (t->which == emApp ==> separated_Mapp(keys, len, t->mApp)) && 
  @      (t->which == emLam ==> separated_Mlam(keys, len, t->mLam)) && 
  @      (t->which == emWit ==> separated_Mwit(keys, len, t->mWit)) && 
  @      (t->which == emAbs ==> separated_Mabs(keys, len, t->mAbs)) &&
  @      (t->which == emTup ==> separated_Mtup(keys, len, t->mTup)) && 
  @      (t->which == emPi1 ==> separated_Mpi1(keys, len, t->mPi1)) && 
  @      (t->which == emPi2 ==> separated_Mpi2(keys, len, t->mPi2)) && 
  @      (t->which == emIn1 ==> separated_Min1(keys, len, t->mIn1)) && 
  @      (t->which == emIn2 ==> separated_Min2(keys, len, t->mIn2)) && 
  @      (t->which == emCas ==> separated_Mcas(keys, len, t->mCas)) && 
  @      (t->which == emEps ==> separated_Meps(keys, len, t->mEps));
  @
  @ predicate separated_Kind{L}(char** keys, size_t len, Kind *t);
  @
  @ predicate separated_Kpi{L}(char** keys, size_t len, Kpi m) = 
  @     \separated(keys+(0..(sizeof(char*)*len)-1), m.name) &&
  @     separated_Type(keys, len, m.type) &&
  @     separated_Kind(keys, len, m.kind);

  @ axiom separated_Kind_def{L}:
  @     \forall char** keys, size_t len, Kind *t;
  @      separated_Kind(keys, len, t) <==>
  @       \separated(keys+(0..(sizeof(char*)*len)-1), t) &&
  @       (t->which == eKstar ==> \true) &&
  @       (t->which == eKpi ==> separated_Kpi(keys, len, t->kPi));

  @  predicate separated_Bind{L}(char** keys, size_t len, Bind t);
  @
  @  predicate separated_Any{L}(char** keys, size_t len, Any *t) =
  @    \separated(keys+(0..(sizeof(char*)*len)-1), t) &&
  @    (t->which == eAnyType ==> separated_Type(keys, len, t->mType)) &&
  @    (t->which == eAnyTerm ==> separated_Term(keys, len, t->mTerm)) &&
  @    (t->which == eAnyKind ==> separated_Kind(keys, len, t->mKind)) &&
  @    (t->which == eAnyName ==> separated_Name(keys, len, t->mName)) &&
  @    (t->which == eAnyBind ==> separated_Bind(keys, len, t->mBind));
  @
  @
  @ }
  @*/

/*@ axiomatic destruct_validity { 
  @    lemma valid_Term_to_valid_App:
  @      \forall Term *t;  valid_Term_ptr(t) && t->which == emApp ==>
  @          valid_Term_ptr(t->mApp.term1);
  @ }
  @*/


/*@ requires valid_Term_ptr(t);
  @ assigns \nothing;
  @ ensures valid_Any(\result); 
  @*/
Any anyOfTerm(Term *t);

/*@ requires valid_Type_ptr(t);
  @ assigns \nothing;
  @ ensures valid_Any(\result); 
  @*/ 
Any anyOfType(Type *t);

/*@ requires valid_Kind_ptr(t);
  @ assigns \nothing;
  @ ensures valid_Any(\result);
  @*/
Any anyOfKind(Kind *t);

/*@ requires valid_Name(t);
  @ assigns \nothing;
  @ ensures valid_Any(\result); 
  @*/ 
Any anyOfName(Name t);

/*@ requires valid_Bind(b);
  @ assigns \nothing;
  @ ensures valid_Any(\result); 
  @ */ 
Any anyOfBind(const Bind b);

/*@ requires valid_Name(n);
  @ requires valid_string((char*)errorBuffer) && 0 <= errorNum < 2147483647;
  @ requires strlen((char*)errorBuffer) < L_BUFFER_LEN;
  @ requires is_allocable((size_t)sizeof(Term),(size_t)1);
  @ assigns errorBuffer[0..L_BUFFER_LEN-1], errorNum;
  @ assigns __allocated[0..MAX_ALLOCS], __size_allocated, __num_allocated;
  @ ensures valid_Term_ptr(\result); 
  @ ensures \at(__num_allocated,Pre)+1==\at(__num_allocated,Post);
  @ ensures \forall size_t n,s; is_allocable{Pre}(n,s) <==> is_allocable{Post}((size_t)(n-sizeof(Term)),(size_t)(s-1));
  @*/
Term *mMvar(Name n);

/*@ requires valid_Name(n);
  @ requires valid_string((char*)errorBuffer) && 0 <= errorNum < 2147483647;
  @ requires strlen((char*)errorBuffer) < L_BUFFER_LEN;
  @ requires is_allocable((size_t)sizeof(Type),(size_t)1);
  @ assigns errorBuffer[0..L_BUFFER_LEN-1], errorNum;
  @ assigns __allocated[0..MAX_ALLOCS], __size_allocated, __num_allocated;
  @ ensures \at(__num_allocated,Pre)+1==\at(__num_allocated,Post);
  @ ensures valid_Type_ptr(\result); 
  @*/
Type *tTvar(Name n);

/*@ requires valid_Name(n) && valid_Type_ptr(t) && valid_Any_ptr(a);
  @ assigns \result;
  @ ensures valid_Bind(\result); 
  @ ensures \forall size_t n,s; is_allocable{Pre}(n,s) <==> is_allocable{Post}(n,s);
  @*/
Bind mBind(Name n, Type *t, Any *a);

#endif
