/****************************************************************************
 * Copyright (C) 2016 by University of Warsaw                               *
 *                                                                          *
 * This file is part of IntuitionTC.                                        *
 *                                                                          *
 ****************************************************************************/
#include <stdbool.h>
#include <assert.h>
#include "envmap_base.h"
#include "alloc.h"
#include "gen_spec_checker.h"
#include "errorMessages.h"
#include "namemap.h"

/*@ predicate not_modified_charr{L,M}(char** p, size_t size) =
  @    \forall size_t i; 0 <= i < size ==> \at(p[i],L) == \at(p[i],M);
  @ predicate not_modified_anyr{L,M}(Any** p, size_t size) =
  @    \forall size_t i; 0 <= i < size ==> \at(p[i],L) == \at(p[i],M);
  @ 
  @*/


const size_t DEFAULT_SIZE = 100;


/*@
  @ requires \valid(dest+(0..charr_bufsize));
  @ requires \valid_read(src+(0..size-1));
  @ requires src==\null <==> dest==\null;
  @ requires size <= charr_bufsize;
  @ requires \separated(dest+(0..(size_t)(charr_bufsize-1)), src+(0..charr_bufsize-1));
  @ requires 0 <= size < UINT32_MAX;
  @ requires \separated(__allocated+(0..MAX_ALLOCS-1),(char*)dest+(0..sizeof(char*)*charr_bufsize-1));
  @ requires is_allocated_size((char*)src, (size_t)(charr_bufsize*sizeof(char*))) &&
  @                is_allocated_size((char*)dest, (size_t)(charr_bufsize*sizeof(char*)));
  @ requires \valid{Pre}(src+(0..(size_t)(charr_bufsize-1))) && \valid{Pre}(dest+(0..(size_t)(charr_bufsize-1)));
  @ requires 
  @      \separated(__allocated+(0..MAX_ALLOCS-1),(char*)src+(0..sizeof(char*)*charr_bufsize-1)) ==>
  @      \separated(__allocated+(0..MAX_ALLOCS-1),(char*)dest+(0..sizeof(char*)*charr_bufsize-1));
  @ requires valid_anyarray{Pre}(src, charr_bufsize, size);
  @ assigns dest[0..size-1];
  @ exits \false;
  @ ensures \forall size_t i; 0 <= i < size ==> \at(dest[i],Post)==\at(src[i],Pre);
  @ ensures \forall char* p, size_t i; is_allocated_size{Pre}(p, i) ==> is_allocated_size{Post}(p,i);
  @ ensures \forall char* p, size_t i; \valid{Pre}(p+(0..i)) ==> \valid{Post}(p+(0..i));
  @ ensures \forall char** p, size_t i; \valid{Pre}(p+(0..i)) ==> \valid{Post}(p+(0..i));
  @ ensures \forall size_t i; 0 <= i < MAX_ALLOCS ==> \at(__allocated[i],Pre) == \at(__allocated[i],Post);
  @ ensures \forall char* p,*q, size_t n,m; \separated(\at(p,Pre)+(0..n-1),\at(q,Pre)+(0..m-1)) ==>
  @                                         \separated(\at(p,Post)+(0..n-1),\at(q,Post)+(0..m-1));
  @ ensures \forall char** p,**q, size_t n,m; \separated(\at(p,Pre)+(0..n-1),\at(q,Pre)+(0..m-1)) ==>
  @                                         \separated(\at(p,Post)+(0..n-1),\at(q,Post)+(0..m-1));
  @ ensures valid_anyarray{Post}(src, charr_bufsize, size);
  @ ensures valid_anyarray{Post}(dest, charr_bufsize, size);
  @ ensures \at(__num_allocated,Pre) == \at(__num_allocated,Post);
  @ ensures \at(__size_allocated,Pre) == \at(__size_allocated,Post);
  @*/
void memcpy_any(Any** dest, Any** src, size_t size) {
  size_t i = 0;
  /*@ loop invariant \forall size_t j; 0 <= j < i ==> dest[j]==\at(src[j],LoopEntry);
    @ loop invariant \forall size_t j; 0 <= j < size ==> src[j]==\at(src[j],LoopEntry);
    @ loop invariant size <= charr_bufsize;
    @ loop invariant \valid(dest+(0..size-1));
    @ loop invariant \valid_read(src+(0..size-1));
    @ loop invariant src==\null <==> dest==\null;
    @ loop invariant \separated(dest+(0..charr_bufsize-1), src+(0..charr_bufsize-1));
    @ loop invariant \separated(__allocated+(0..MAX_ALLOCS-1),(char*)dest+(0..sizeof(char*)*charr_bufsize-1));
    @ loop invariant 0 <= i <= size;
    @ loop invariant is_allocated_size((char*)src, (size_t)(charr_bufsize*sizeof(char*))) &&
    @                is_allocated_size((char*)dest, (size_t)(charr_bufsize*sizeof(char*)));
    @ loop invariant \forall char* p, size_t i; \valid{Pre}(p+(0..i)) ==> \valid{Here}(p+(0..i));
    @ loop invariant \forall char** p, size_t i; \valid{Pre}(p+(0..i)) ==> \valid{Here}(p+(0..i));
    @ loop invariant \forall size_t i; 0 <= i < MAX_ALLOCS ==> \at(__allocated[i],LoopEntry) == __allocated[i];
    @ loop invariant \forall char* p,*q, size_t n,m; \separated(\at(p,LoopEntry)+(0..n-1),\at(q,LoopEntry)+(0..m-1)) ==>
    @                                                \separated(\at(p,Here)+(0..n-1),\at(q,Here)+(0..m-1));
    @ loop invariant \forall char** p,**q, size_t n,m; \separated(\at(p,LoopEntry)+(0..n-1),\at(q,LoopEntry)+(0..m-1)) ==>
    @                                                \separated(\at(p,Here)+(0..n-1),\at(q,Here)+(0..m-1));
    @ loop invariant valid_anyarray(dest, charr_bufsize, i);
    @ loop invariant \at(__num_allocated,Pre) == __num_allocated;
    @ loop invariant \at(__size_allocated,Pre) == __size_allocated;
    @ loop invariant valid_anycontent(src,size);
    @ loop assigns i,dest[0..size-1];
    @*/
  while (i<size) {
    dest[i]=src[i];
    //@ assert i<size;
    i++;
    //@ assert \forall size_t j; 0 <= j <= i ==> (0 <= j < i || j == i);
  }
  //@ assert size==i;
}


envmap envmap_new() {
  	char **keys = calloc(DEFAULT_SIZE, sizeof(*keys));
	Any **vals = calloc(DEFAULT_SIZE, sizeof(*vals));
        if (keys == NULL && vals != NULL) {
          free(vals);
          return (envmap){
            .size = 0, .bufsize = 0,
              .keys = NULL, .vals = NULL};
        } else {
          if (keys != NULL && vals == NULL) {
            free(keys);
            return (envmap){
              .size = 0, .bufsize = 0,
                .keys = NULL, .vals = NULL};
          } else {
            return (envmap){
              .size = 0, .bufsize = (keys)?DEFAULT_SIZE:0,
                .keys = keys, .vals = vals};
          }
        }
}

void envmap_free(envmap n) {
	free(n.keys);
	free(n.vals);
	n.size = n.bufsize = 0;
}

envmap *envmap_insert(envmap *n, Name elem, Any val) {
        char **tmp=NULL;
        Any **tmpa=NULL;
        unsigned int nsize = n->bufsize * 2;
        if (n->size == n->bufsize) {
		tmp=calloc(nsize, sizeof(char*));
		if (tmp==NULL) return NULL;
                //@ ghost charr_bufsize = n->bufsize;
                memcpy_charr(tmp, n->keys, n->size);
                free(n->keys);
                n->keys=tmp;
		tmpa=calloc(nsize, sizeof(Any*));
                if (tmpa==NULL) return NULL;
                //@ assert well_separated_arrays(n->keys, tmpa, n->bufsize, (size_t)0);
                //@ ghost charr_bufsize = n->bufsize;
                memcpy_any(tmpa, n->vals, n->size);
                free(n->vals);
                n->vals=tmpa;
                n->bufsize=nsize;
                //@ assert well_separated_envmap(n->keys, n->size, n->bufsize, n);
                //@ assert well_separated_any(n->vals, n->size, n->bufsize, n);
                //@ assert is_allocated_size((char*)n, (size_t)sizeof(namemap));
                //@ assert well_separated_arrays(n->keys, n->vals, n->bufsize, n->size);
                //@ assert \forall size_t i; 0 <= i < n->size ==> separated_Any(n->keys, n->bufsize, n->vals[i]);
	}
        //@ assert is_allocated_size((char*)n, (size_t)sizeof(namemap));
        //@ assert n->size <= n->bufsize;
        //@ assert well_separated_envmap(n->keys, n->size, n->bufsize, n);
        //@ assert well_separated_any(n->vals, n->size, n->bufsize, n);
        //@ assert well_separated_arrays(n->keys, n->vals, n->bufsize, n->size);
        //@ assert \forall size_t i; 0 <= i < n->size ==> separated_Any(n->keys, n->bufsize, n->vals[i]);
	n->keys[n->size] = elem;
        //@ assert well_separated_envmap(n->keys, n->size, n->bufsize, n);
        //@ assert well_separated_any(n->vals, n->size, n->bufsize, n);
        //@ assert well_separated_arrays(n->keys, n->vals, n->bufsize, n->size);
        //@ assert \forall size_t i; 0 <= i < n->size ==> separated_Any(n->keys, n->bufsize, n->vals[i]);
	n->vals[n->size] = makeAny(val);
        if (n->vals[n->size] == NULL) {
          return NULL;
        }
        //@ assert well_separated_envmap(n->keys, n->size, n->bufsize, n);
        //@ assert well_separated_any(n->vals, n->size, n->bufsize, n);
        //@ assert well_separated_arrays(n->keys, n->vals, n->bufsize, n->size);
        //@ assert \forall size_t i; 0 <= i < n->size ==> separated_Any(n->keys, n->bufsize, n->vals[i]);
	n->size++;
        //@ assert valid_string(n->keys[n->size-1]);
        //@ assert valid_sizes(n->size, n->bufsize);
        //@ assert valid_charray(n->keys, n->bufsize, n->size);
        //@ assert valid_anyarray(n->vals, n->bufsize, n->size);
        //@ assert \forall size_t i; 0 <= i < n->size ==> separated_Any(n->keys, n->bufsize, n->vals[i]);
	return n;
}

Any *envmap_lookup(const envmap n, const Name elem) {
	/* NULL == Nothing */
	int i;
        /*@
          @ loop invariant valid_Name(elem);
          @ loop invariant valid_envmap(n);
          @ loop invariant 0 <= i <= n.size;
          @ loop invariant !IsInEnv(n, i, elem);
          @ loop assigns i;
          @ loop variant n.size - i;
          @*/
	for (i = 0; i < n.size; ++i) {
          //@ assert valid_Any_ptr(n.vals[i]);
		if (strcmp(elem, n.keys[i]) == 0)
			return n.vals[i];
        }
	return NULL;
}

/* shallow */
envmap* envmap_copy(const envmap in) {
  char **keys;
  Any  **vals;
  envmap* res = calloc(1, sizeof(envmap));
  if (res == NULL) {
    //@ for alloc_new: assert \false;
    return NULL;
  }
 BB:
  //@ for alloc_new: assert is_allocable{Pre}((size_t)(in.bufsize*sizeof(char*)+in.bufsize*sizeof(Any*)),(size_t)2);
  //@ assert valid_anyarray(in.vals, in.bufsize, in.size);
  //@ assert valid_charray(in.keys, in.bufsize, in.size);
  //@ assert correct_allocation((size_t)0);
  keys = calloc(in.bufsize, sizeof(char*));
  if (keys == NULL) {
    // @ for alloc_new: assert \false;
    free(res);
    //@ for alloc_no: assert allocation_footprint{Pre,Here}((size_t)0, (size_t)0);
    return NULL;
  }
  //@ for alloc_new: assert is_allocable((size_t)(in.bufsize*sizeof(Any*)),(size_t)1);
  //@ assert valid_anyarray(in.vals, in.bufsize, in.size);
  //@ assert valid_charray(in.keys, in.bufsize, in.size);
  //@ assert well_separated_envmap(keys, in.size, in.bufsize, res);
 AA:
  vals = calloc(in.bufsize, sizeof(Any*));
  //@ assert not_modified_charr{AA,Here}(in.keys, in.size);
  //@ assert not_modified_anyr{AA,Here}(in.vals, in.size);
  // @ assert __num_allocated <= MAX_ALLOCS;
  //@ assert valid_anyarray(in.vals, in.bufsize, in.size);
  //@ assert valid_charray(in.keys, in.bufsize, in.size);
  //@ assert well_separated_envmap(keys, in.size, in.bufsize, res);
  if (vals == NULL) {
    //@ for alloc_new: assert \false;
    //@ assert allocation_footprint{Pre,Here}((size_t)((in.bufsize*sizeof(char*))+sizeof(envmap)), (size_t)2);
    free(res);
    //@ for alloc_no: assert \at(__num_allocated,Pre)+1 == __num_allocated;
    free(keys);
    //@ for alloc_new: assert \false;
    //@ for alloc_no: assert allocation_footprint{Pre,Here}((size_t)0, (size_t)0);
    return NULL;
  }
  //@ assert vals!=\null;
  //@ assert keys!=\null;
  //@ ghost charr_bufsize = in.bufsize;
  // @ assert is_allocable{Pre}((size_t)((2*in.bufsize*sizeof(char*))+sizeof(namemap)),(size_t)3);
  //@ assert \separated((char*)keys+(0..(size_t)(in.bufsize*sizeof(char*))-1),(char*)(in.keys)+(0..(size_t)(in.bufsize*sizeof(char*)-1)));
  //@ assert is_allocated_size((char*)keys, (size_t)(in.bufsize*sizeof(char*)));
  //@ assert is_allocated_size((char*)vals, (size_t)(in.bufsize*sizeof(Any*)));
  //@ assert valid_charray(in.keys, in.bufsize, in.size);
  //@ assert well_separated_envmap(keys, in.size, in.bufsize, res);
  //@ assert well_separated_any(vals, (size_t)0, in.bufsize, res);
 A:
  memcpy_charr(keys, in.keys, in.size);
  //@ assert not_modified_charr{A,Here}(in.keys, in.size);
  //@ assert not_modified_anyr{A,Here}(in.vals, in.size);
  //@ ghost charr_bufsize = in.bufsize;
  //@ for alloc_new: assert \separated(vals+(0..(size_t)(in.bufsize-1)),in.vals+(0..(size_t)(in.bufsize-1)));
  //@ assert is_allocated_size((char*)keys, (size_t)(in.bufsize*sizeof(char*)));
  //@ assert is_allocated_size((char*)vals, (size_t)(in.bufsize*sizeof(Any*)));
  //@ assert valid_charray(keys, in.bufsize, in.size);
  //@ assert valid_anyarray(in.vals, in.bufsize, in.size);
  //@ assert well_separated_envmap(keys, in.size, in.bufsize, res);
  //@ assert well_separated_any(vals, in.size, in.bufsize, res);
  memcpy_any(vals, in.vals, in.size);
  //@ assert valid_charray(keys, in.bufsize, in.size);
  //@ assert valid_anyarray(vals, in.bufsize, in.size);
  //@ assert valid_sizes(in.size, in.bufsize);
  //@ assert is_allocated_size((char*)keys, (size_t)(in.bufsize*sizeof(char*)));
  //@ assert is_allocated_size((char*)vals, (size_t)(in.bufsize*sizeof(Any*)));
  //@ assert well_separated_envmap(keys, in.size, in.bufsize, res);
  //@ assert well_separated_any(vals, in.size, in.bufsize, res);
  res->size = in.size;
  res->bufsize = in.bufsize;
  res->keys = keys;
  res->vals = vals;
  //@ assert well_separated_envmap(res->keys, res->size, res->bufsize, res);
  //@ assert well_separated_any(res->vals, res->size, res->bufsize, res);
  //@ assert valid_charray(res->keys, res->bufsize, res->size);
  //@ assert valid_anyarray(res->vals, res->bufsize, res->size);
  //@ assert valid_sizes(res->size, res->bufsize);
  return res;

}



