/****************************************************************************
 * Copyright (C) 2016 by University of Warsaw                               *
 *                                                                          *
 * This file is part of IntuitionTC.                                        *
 *                                                                          *
 ****************************************************************************/
/**
 * @file envmap_base.h
 * @brief Operations on the data structure that represents environments.
 * @author Aleksy Schubert
 * @author Maciej Zielenkiewicz
 */
#ifndef _ENVMAP_BASE_H
#define _ENVMAP_BASE_H  1
#include <string.h>
#include "checker.h"
#ifdef _FRAMAC_
#include "namemap.h"
#endif

/** @brief The representation of the dictionary to hold values wrapped in
 *         #Any structure and available under string keys.
 * 
 *  The i-th entry in #keys represents the i-th key while the i-th entry
 *  in #vals represents the corresponding value.
 */
typedef struct {
  /** @brief The size of the area in the representation which is occupied by
   *  valid keys and values. */
  size_t size;
  /** @brief The size of the area that can hold valid keys and values before
   *  reallocation is done.  */
  size_t bufsize;
  /** @brief The array with the dictionary keys. */
  char **keys;
  /** @brief The array with the dictionary values keys. */
  Any **vals;
} envmap;

/*@
  @  predicate IsInEnv{A}(envmap n, integer size, Name elem) =
  @    \exists integer i; 0 <= i < size && strcmp(elem, n.keys[i]) == 0;
  @
  @  predicate IsInResEnv{A}(namemap n, Name where, Any* what) =
  @    \exists integer i; 0 <= i < n.size && strcmp(where, n.keys[i]) == 0 &&
  @                       \true;
  @
  @*/


/*@
  @ //Both sizes are non-negative and bufsize is not less than size.
  @
  @ predicate valid_anycontent{L}(Any** keys, size_t size) =
  @   \forall size_t i; 0 <= i < size ==> (keys[i] != \null &&
  @                                     valid_Any_ptr(keys[i]));
  @
  @ predicate valid_anyarray{L}(Any** vals, size_t bufsize, size_t size) =
  @   \separated(__allocated+(0..MAX_ALLOCS-1),(char*)vals+(0..bufsize*sizeof(Any*)-1)) &&
  @   vals != \null &&
  @   \valid{L}(vals+(0..bufsize-1)) &&
  @   is_allocated_size{L}((char*)vals, (size_t)(bufsize*sizeof(Any*))) &&
  @   valid_anycontent{L}(vals, size);
  @
  @ predicate well_separated_envmap{L}(char** keys, size_t size, size_t bufsize, envmap *n) =
  @   \separated(((char*)keys)+(0..(bufsize*sizeof(char*))-1), n) &&
  @   \forall size_t i; 0 <= i < size ==> 
  @           \separated(((char*)keys)+(0..(bufsize*sizeof(char*))-1), (keys[i])+(0..\block_length(keys[i])-1)) &&
  @           \separated((keys[i])+(0..\block_length(keys[i])-1),n);
  @
  @
  @ predicate well_separated_any{L}(Any** vals, size_t size, size_t bufsize, envmap *n) =
  @   \separated(((char*)vals)+(0..(bufsize*sizeof(Any*))-1), n) &&
  @   \forall size_t i; 0 <= i < size ==> 
  @           \separated(((char*)vals)+(0..(bufsize*sizeof(Any*))-1), (vals[i])+(0..sizeof(Any)-1)) &&
  @           \separated((vals[i])+(0..\block_length(vals[i])-1),n);
  @
  @
  @ predicate well_separated_arrays{L}(char** keys, Any** vals, size_t bufsize, size_t size) =
  @   \separated(((char*)keys)+(0..(bufsize*sizeof(char*))-1), 
  @              ((char*)vals)+(0..(bufsize*sizeof(Any*))-1)) &&
  @   \forall size_t i; 0 <= i < size ==>
  @             \separated(((char*)keys)+(0..(bufsize*sizeof(char*))-1), 
  @                        ((char*)vals[i])+(0..(sizeof(Any))-1)) &&
  @   \forall size_t i; 0 <= i < size ==>
  @             \separated(((char*)vals)+(0..(bufsize*sizeof(Any*))-1), 
  @                        (keys[i])+(0..\block_length(keys[i])-1));
  @
  @
  @
  @ predicate valid_envmap{L}(envmap n) = 
  @    valid_sizes(n.size, n.bufsize) && (char*)n.keys!= (char*)n.vals &&
  @    valid_charray{L}(n.keys, n.bufsize, n.size) &&
  @    valid_anyarray{L}(n.vals, n.bufsize, n.size);
  @
  @
  @ predicate valid_envmap_ptr{L}(envmap *n) =
  @    n != \null &&
  @    \valid{L}(n) &&
  @    is_allocated_size{L}((char*)n, (size_t)sizeof(envmap)) &&
  @    valid_envmap{L}(*n) &&
  @    well_separated_arrays{L}(n->keys, n->vals, n->bufsize, n->size) &&
  @    well_separated_envmap{L}(n->keys, n->size, n->bufsize, n) &&
  @    well_separated_any{L}(n->vals, n->size, n->bufsize, n) &&
  @    \forall size_t i; 0 <= i < n->size ==> separated_Any(n->keys, n->bufsize, n->vals[i]);
  @*/



/*@ requires default_size_inv(DEFAULT_SIZE) && correct_allocation((size_t)0);
  @ behavior alloc_new:
  @   assumes is_allocable{Pre}((size_t)(DEFAULT_SIZE*sizeof(char*)+DEFAULT_SIZE*sizeof(Any*)),(size_t)2);
  @   assigns __allocated[__num_allocated..__num_allocated+1], __size_allocated, __num_allocated;
  @   ensures valid_envmap(\result);
  @   ensures allocation_footprint{Pre,Post}((size_t)(DEFAULT_SIZE*sizeof(char*)+DEFAULT_SIZE*sizeof(Any*)),(size_t)2);
  @ behavior alloc_no: 
  @   assumes !is_allocable{Pre}((size_t)(DEFAULT_SIZE*sizeof(char*)+DEFAULT_SIZE*sizeof(Any*)),(size_t)2);
  @   assigns __allocated[__num_allocated];
  @   ensures \result.size==0 && \result.bufsize==0 && \result.keys == \null && \result.vals == \null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
envmap envmap_new();

/*@ requires valid_envmap(n) && correct_allocation((size_t)0);
  @ assigns __allocated[0..MAX_ALLOCS], __size_allocated, __num_allocated;
  @ ensures \at(__num_allocated,Post) == \at(__num_allocated,Pre) - 2;
  @ ensures \at(__size_allocated,Post) == \at(__size_allocated,Pre)  - 
  @                                n.bufsize*sizeof(char*)+n.bufsize*sizeof(Any*);
  @*/
void envmap_free(envmap n); /* shallow */

/*@ requires valid_envmap_ptr(n) &&
  @          valid_Name(elem) && valid_Any(val);
  @ requires \separated(n,elem+(0..\block_length(elem)-1)) &&
  @          \separated((char*)(n->vals)+(0..sizeof(char*)*n->bufsize-1),
                        elem+(0..\block_length(elem)-1)) &&
  @          \separated(n->keys+(0..n->bufsize-1),
                        elem+(0..\block_length(elem)-1));
  @ requires 0 < n->bufsize*sizeof(char*)+n->bufsize*sizeof(Any*) < UINT32_MAX;
  @ requires 0 <= n->bufsize <= UINT32_MAX;
  @ requires (n->size == n->bufsize) ==>
  @             is_allocable((size_t)(n->bufsize*sizeof(char*)+n->bufsize*sizeof(Any*)+sizeof(Any)),(size_t)2);
  @ requires (n->size != n->bufsize) ==>
  @             is_allocable((size_t)(sizeof(Any)),(size_t)1);
  @ requires valid_Name(elem) && valid_Any(val);
  @ requires \separated(((char*)(n->keys)+(0..(n->bufsize*sizeof(char*))-1)), 
  @              elem+(0..\block_length(elem)-1));
  @ ensures \result!=\null ==> valid_envmap_ptr(\result);
  @*/
envmap *envmap_insert(envmap *n, Name elem, Any val);

/*@ 
  @ requires valid_envmap(n) && valid_Name(elem) && correct_allocation((size_t)0);
  @  assigns \nothing;
  @ 
  @ behavior found:
  @  assumes IsInEnv(n, n.size, elem);
  @  ensures \result!= \null && valid_Any_ptr(\result); // TODO  && IsInResEnv(n, elem, \null); // *\result);
  @ behavior non_found:
  @  assumes !IsInEnv(n, n.size, elem);
  @  ensures \result==\null;
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Any *envmap_lookup(const envmap n, const Name elem);

/*@ 
  @ requires default_size_inv(DEFAULT_SIZE) && valid_envmap(in) && correct_allocation((size_t)0);
  @ requires 0 <= in.bufsize*sizeof(char*) + in.bufsize*sizeof(Any*) + sizeof(envmap) <= UINT32_MAX;
  @ requires \forall size_t i; 0 <= i < in.size ==>
  @             \exists integer j; j>0 && depth_Any((in.vals)[i])==j;
  @
  @ behavior alloc_new:
  @   assumes is_allocable{Pre}((size_t)(in.bufsize*sizeof(char*)+in.bufsize*sizeof(Any*)+sizeof(envmap)),(size_t)3);
  @   assigns __allocated[__num_allocated..__num_allocated+1], __size_allocated, __num_allocated;
  @   ensures valid_envmap_ptr(\result);
  @   ensures allocation_footprint{Pre,Post}((size_t)(in.bufsize*sizeof(char*)+in.bufsize*sizeof(Any*)+sizeof(envmap)), (size_t)3);
  @ behavior alloc_no: 
  @   assumes !is_allocable{Pre}((size_t)(in.bufsize*sizeof(char*)+in.bufsize*sizeof(Any*)+sizeof(envmap)),(size_t)3);
  @   assumes __num_allocated <= UINT32_MAX - 1;
  @   assigns __allocated[__num_allocated];
  @   ensures \result == \null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0, (size_t)0);
  @
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
envmap* envmap_copy(const envmap in); /* shallow */

#endif
