/****************************************************************************
 * Copyright (C) 2016 by University of Warsaw                               *
 *                                                                          *
 * This file is part of IntuitionTC.                                        *
 *                                                                          *
 ****************************************************************************/
/**
 * @file namemap.h
 * @brief Operations on the data structure that represents dictionaries of
 *        names.
 * @author Aleksy Schubert
 * @author Maciej Zielenkiewicz
 */
#ifndef _NAMEMAP_H
#define _NAMEMAP_H

#include <string.h>
#include "checker.h"
#ifdef _FRAMAC_
#include "libc/__fc_namemap.h"
#else

/** @brief The representation of the dictionary to hold strings under
 *  string keys.
 * 
 *  The i-th entry in #keys represents the i-th key while the i-th entry
 *  in #vals represents the corresponding value.
 */
typedef struct {
  /** @brief The size of the area in the representation which is occupied by
   *  valid keys and values. */
  size_t size;
  /** @brief The size of the area that can hold valid keys and values before
   *  reallocation is done.  */
  size_t bufsize;
  /** @brief The array with the dictionary keys. */
  char **keys;
  /** @brief The array with the dictionary values keys. */
  char **vals;
} namemap;
#endif


/*@
  @  predicate IsIn{A}(namemap n, integer size, Name elem) =
  @    \exists integer i; 0 <= i < size && strcmp(elem, n.keys[i]) == 0;
  @  predicate IsInRes{A}(namemap n, Name where, Name what) =
  @    \exists integer i; 0 <= i < n.size && strcmp(where, n.keys[i]) == 0 &&
  @                       n.vals[i] == what;
  @
  @*/

/*@
  @ //Both sizes are non-negative and bufsize is not less than size.
  @ predicate valid_sizes(size_t size, size_t bufsize) =
  @   0<= size <= UINT32_MAX && 0 <= 2*bufsize*sizeof(char*) <= UINT32_MAX && bufsize >= size && 0 < bufsize;
  @ 
  @ predicate valid_content{L}(char** keys, size_t size) =
  @   \forall int i; 0 <= i < size ==> valid_string(keys[i]);
  @
  @ predicate valid_charray{L}(char** keys, size_t bufsize, size_t size) =
  @   \separated(__allocated+(0..MAX_ALLOCS-1),
                 (char*)keys+(0..bufsize*sizeof(char*)-1)) &&
  @   keys != \null &&
  @   \valid{L}(keys+(0..bufsize-1)) &&
  @   is_allocated_size{L}((char*)keys, (size_t)(bufsize*sizeof(char*))) &&
  @   valid_content{L}(keys, size);
  @
  @ predicate well_separated{L}(char** keys, size_t size, namemap *n) =
  @   \separated(keys+(0..size-1), n);
  @
  @ predicate well_separated_arrays{L}(char** keys, char** vals, size_t size) =
  @   \separated(keys+(0..(size-1)), vals+(0..(size-1)));
  @
  @ predicate valid_namemap{L}(namemap n) = 
  @    valid_sizes(n.size, n.bufsize) && n.keys!= n.vals &&
  @    valid_charray{L}(n.keys, n.bufsize, n.size) &&
  @    valid_charray{L}(n.vals, n.bufsize, n.size);
  @
  @ predicate valid_namemap_ptr{L}(namemap *n) =
  @    n!=\null &&
  @    \valid{L}(n) &&
  @    is_allocated_size{L}((char*)n, (size_t)sizeof(namemap)) &&
  @    valid_namemap{L}(*n) &&
  @    well_separated_arrays{L}(n->keys, n->vals, n->bufsize) &&
  @    well_separated{L}(n->keys, n->bufsize, n) &&
  @    well_separated{L}(n->vals, n->bufsize, n);
  @*/

//@ ghost size_t charr_bufsize;

/*@
  @ requires \valid(dest+(0..charr_bufsize));
  @ requires \valid_read(src+(0..size-1));
  @ requires src==\null <==> dest==\null;
  @ requires size <= charr_bufsize;
  @ requires \separated(dest+(0..(size_t)(charr_bufsize-1)), src+(0..charr_bufsize-1));
  @ requires 0 <= size < UINT32_MAX;
  @ requires \separated(__allocated+(0..MAX_ALLOCS-1),(char*)dest+(0..sizeof(char*)*charr_bufsize-1));
  @ requires is_allocated_size((char*)src, (size_t)(charr_bufsize*sizeof(char*))) &&
  @                is_allocated_size((char*)dest, (size_t)(charr_bufsize*sizeof(char*)));
  @ requires \valid{Pre}(src+(0..(size_t)(charr_bufsize-1))) && \valid{Pre}(dest+(0..(size_t)(charr_bufsize-1)));
  @ requires 
  @      \separated(__allocated+(0..MAX_ALLOCS-1),(char*)src+(0..sizeof(char*)*charr_bufsize-1)) ==>
  @      \separated(__allocated+(0..MAX_ALLOCS-1),(char*)dest+(0..sizeof(char*)*charr_bufsize-1));
  @ requires valid_charray{Pre}(src, charr_bufsize, size);
  @ assigns dest[0..size-1];
  @ exits \false;
  @ ensures \forall size_t i; 0 <= i < size ==> \at(dest[i],Post)==\at(src[i],Pre);
  @ ensures \forall char* p, size_t i; is_allocated_size{Pre}(p, i) ==> is_allocated_size{Post}(p,i);
  @ ensures \forall char* p, size_t i; \valid{Pre}(p+(0..i)) ==> \valid{Post}(p+(0..i));
  @ ensures \forall char** p, size_t i; \valid{Pre}(p+(0..i)) ==> \valid{Post}(p+(0..i));
  @ ensures \forall size_t i; 0 <= i < MAX_ALLOCS ==> \at(__allocated[i],Pre) == \at(__allocated[i],Post);
  @ ensures \forall char* p,*q, size_t n,m; \separated(\at(p,Pre)+(0..n-1),\at(q,Pre)+(0..m-1)) ==>
  @                                         \separated(\at(p,Post)+(0..n-1),\at(q,Post)+(0..m-1));
  @ ensures \forall char** p,**q, size_t n,m; \separated(\at(p,Pre)+(0..n-1),\at(q,Pre)+(0..m-1)) ==>
  @                                         \separated(\at(p,Post)+(0..n-1),\at(q,Post)+(0..m-1));
  @ ensures valid_charray{Post}(src, charr_bufsize, size);
  @ ensures valid_charray{Post}(dest, charr_bufsize, size);
  @ ensures \at(__num_allocated,Pre) == \at(__num_allocated,Post);
  @ ensures \at(__size_allocated,Pre) == \at(__size_allocated,Post);
  @*/
void memcpy_charr(char** dest, char** src, size_t size);


/*@ requires default_size_inv(DEFAULT_SIZE) && correct_allocation((size_t)0);
  @ requires 0 <= 2*DEFAULT_SIZE*sizeof(char*) <= UINT32_MAX;
  @ behavior alloc_new:
  @   assumes is_allocable{Pre}((size_t)(2*DEFAULT_SIZE*sizeof(char*)),(size_t)2);
  @   assigns __allocated[__num_allocated..__num_allocated+1], __size_allocated, __num_allocated;
  @   ensures valid_namemap(\result);
  @   ensures allocation_footprint{Pre,Post}((size_t)(2*DEFAULT_SIZE*sizeof(char*)),(size_t)2);
  @ behavior alloc_no: 
  @   assumes !is_allocable{Pre}((size_t)(2*DEFAULT_SIZE*sizeof(char*)),(size_t)2);
  @   assigns __allocated[__num_allocated];
  @   ensures \result.size==0 && \result.bufsize==0 && \result.keys == \null && \result.vals == \null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0,(size_t)0);
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
namemap namemap_new();

/*@ requires valid_namemap(n) && correct_allocation((size_t)0);
  @ assigns __allocated[0..MAX_ALLOCS], __size_allocated, __num_allocated;
  @ ensures \at(__num_allocated,Post) == \at(__num_allocated,Pre) - 2;
  @ ensures \at(__size_allocated,Post) == \at(__size_allocated,Pre)  - 
  @                                2*n.bufsize*sizeof(char*);
  @*/
void namemap_free(namemap n); /* shallow */

/*@ requires valid_namemap_ptr(n) && correct_allocation((size_t)0);
  @ requires 0 < n->bufsize < UINT32_MAX;
  @ requires 0 <= 2*n->bufsize*sizeof(char*) <= UINT32_MAX;
  @ requires (n->size == n->bufsize) ==>
  @             is_allocable((size_t)(2*n->bufsize*sizeof(char*)),(size_t)1);
  @ requires valid_string(elem) && valid_string(val);
  @ ensures \result!=\null ==> valid_namemap_ptr(\result);
  @*/
namemap *namemap_insert(namemap *n, Name elem, Name val);

/*@ 
  @ requires valid_namemap(n) && valid_Name(elem) && correct_allocation((size_t)0);
  @  assigns \nothing;
  @ 
  @ behavior found:
  @  assumes IsIn(n, n.size, elem);
  @  ensures \result!= \null && IsInRes(n, elem, \result);
  @ behavior non_found:
  @  assumes !IsIn(n, n.size, elem);
  @  ensures \result==\null;
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
Name namemap_lookup(const namemap n, const Name elem);

/*@ 
  @ requires default_size_inv(DEFAULT_SIZE) && valid_namemap(in) && correct_allocation((size_t)0);
  @ requires 0 <= (2*in.bufsize*sizeof(char*)) + sizeof(namemap) <= UINT32_MAX;
  @
  @ behavior alloc_new:
  @   assumes is_allocable{Pre}((size_t)((2*in.bufsize*sizeof(char*))+sizeof(namemap)),(size_t)3);
  @   assigns __allocated[__num_allocated..__num_allocated+1], __size_allocated, __num_allocated;
  @   ensures valid_namemap_ptr(\result); 
  @   ensures allocation_footprint{Pre,Post}((size_t)((2*in.bufsize*sizeof(char*))+sizeof(namemap)), (size_t)3);
  @
  @ behavior alloc_no:
  @   assumes !is_allocable{Pre}((size_t)((2*in.bufsize*sizeof(char*))+sizeof(namemap)),(size_t)3);
  @   assumes __num_allocated <= UINT32_MAX - 1;
  @   assigns __allocated[__num_allocated];
  @   ensures \result == \null;
  @   ensures allocation_footprint{Pre,Post}((size_t)0, (size_t)0);
  @
  @ complete behaviors;
  @ disjoint behaviors;
  @*/
namemap* namemap_copy(const namemap in);

#endif
