/****************************************************************************
 * Copyright (C) 2016 by University of Warsaw                               *
 *                                                                          *
 * This file is part of IntuitionTC.                                        *
 *                                                                          *
 ****************************************************************************/
#include <stdbool.h>
#include <string.h>
#include <assert.h>
#include "typechecker.h"
#include "alloc.h"
#include "free.h"
#include "gen_spec_checker.h"
#include "errorMessages.h"

//! FILE GenChecker.hs



/*@ requires valid_envmap(n) && valid_Name(elem);
  @ requires errorBufferValid(0) && errorNum < 2147483646;
  @ assigns errorBuffer[0..L_BUFFER_LEN-1], errorNum;
  @ ensures valid_string((char*)errorBuffer);
  @*/
//! 353 358
//! -- lookupTvar :: Env -> Name -> CM Kind
//! lookupTvar :: MonadError String m => Env -> Name -> m Kind
//! lookupTvar env n = case lookupVar n env of
//!   Just (Right k) -> return k
//!   Just (Left _) -> reportError ["Variable", n,"is an object variable"]
//!   Nothing -> reportError ["Undeclared type variable",n,"\nenv=",show env]
Kind *envmap_lookupTvar(const envmap n, const Name elem) {
  const Any *const a = envmap_lookup(n, elem);
  if (a == NULL) {
    reportError(messages[ENV_MSG_UNDECLARED_TVAR]);
    //FIXME Nothing -> reportError ["Undeclared type variable",n,"\nenv=",show env]
    return NULL; 
  } else if (a->which == eAnyKind) {
    return a->mKind;
  } else {
    reportError(messages[ENV_MSG_OBJECT_VAR]);
    //FIXME Just (Left _) -> reportError ["Variable", n,"is an object variable"]
    return NULL; 
  }
}

/*@ requires valid_envmap(n) && valid_Name(elem);
  @ requires errorBufferValid(0) && errorNum < 2147483646;
  @ assigns errorBuffer[0..L_BUFFER_LEN-1], errorNum;
  @ ensures valid_string((char*)errorBuffer);
  @*/
//! 360 364
//! lookupMvar :: MonadError String m => Env -> Name -> m Type
//! lookupMvar env n = case lookupVar n env of
//!   Just (Left t) -> return t
//!   Just (Right _) -> reportError ["Variable", n,"is a type variable"]
//!   Nothing -> reportError ["Undeclared mvar",n,"\nenv=",show env]
Type *envmap_lookupMvar(const envmap n, const Name elem) {
  const Any *const a = envmap_lookup(n, elem);
  if (a == NULL) {
    reportError(messages[ENV_MSG_UNDECLARED_MVAR]);
    //FIXME Nothing -> reportError ["Undeclared mvar",n,"\nenv=",show env]
    return NULL; 
  } else if (a->which == eAnyType) {
    return a->mType;
  } else {
    reportError(messages[ENV_MSG_TVAR]);
    //FIXME Just (Right _) -> reportError ["Variable", n,"is a type variable"]
    return NULL; 
  }
}

/*@
  @   requires valid_Name(n);
  @   requires valid_Type_ptr(t);
  @   requires valid_envmap_ptr(env);
  @   assigns \nothing;
  @*/
/* FIXME: envmap is modified */
//! 366 371
//! insertMVar :: MonadError String m => Name -> Type -> Env -> m Env
//! insertMVar n t env = do
//!   checkType env t Kstar -- env |- t : *
//!   case lookup n env of
//!     Nothing -> return $ (n,Left t):env
//!     Just _ -> reportError ["Name",n,"already defined"]
envmap *envmap_insertMVar(Name n, Type *t, envmap *env) {
  Kind k_kStar = (Kind) {.which = eKstar, .kStar = { .nothing = 0 }};
  if (!checkType(*env, t, &k_kStar)) 
    return NULL;
  if (envmap_lookup(*env, n) != NULL) {
    reportError(messages[ENV_MSG_TNAME_DEFINED]);
    // FIXME Just _ -> reportError ["Name",n,"already defined"]
    return NULL;
  }
  return envmap_insert(env, n, anyOfType(t));
}

/* envmap is modified */
//! 366 371
//! insertMVar :: MonadError String m => Name -> Type -> Env -> m Env
//! insertMVar n t env = do
//!   checkType env t Kstar -- env |- t : *
//!   case lookup n env of
//!     Nothing -> return $ (n,Left t):env
//!     Just _ -> reportError ["Name",n,"already defined"]
envmap *envmap_insertTVar(Name n, Kind *k, envmap *env){
  if (!checkKind(*env, *k)) 
    return NULL;
  if (envmap_lookup(*env, n) != NULL) {
    reportError(messages[ENV_MSG_ONAME_DEFINED]);
    // FIXME Just _ -> reportError ["Name",n,"already defined"]
    return NULL;
  }
  return envmap_insert(env, n, anyOfKind(k));
}


/*@ requires valid_envmap(env);
  @ requires valid_Kind(k);
  @ requires valid_string((char*)errorBuffer) && 0 <= errorNum < 2147483647;
  @ requires strlen((char*)errorBuffer) < L_BUFFER_LEN;
  @ requires is_allocable(2*DEFAULT_SIZE*sizeof(char*),(size_t)2);
  @ assigns __allocated[0..MAX_ALLOCS], __size_allocated, __num_allocated;
  @ assigns errorBuffer[0..L_BUFFER_LEN-1], errorNum;
  @*/
//! 383 388
//! -- | Check if a kind is well-formed wrt an environment
//! checkKind :: MonadError String m =>  Env -> Kind -> m ()
//! checkKind env Kstar = return ()
//! checkKind env (Kpi n t k) = do
//!   env' <- insertMVar n t env
//!   checkKind env' k
bool checkKind(const envmap env, Kind k) {
  if (k.which == eKstar)
    return true;

  bool res = true;
  envmap* env_ = envmap_copy(env);
  if (env_ == NULL) return false;
  if (envmap_insertMVar(k.kPi.name, k.kPi.type, env_) == NULL) {
    reportError(messages[ENV_MSG_CANNOT_INSERT_VARIABLE]);
    res = false;
  } else {
    res = checkKind(*env_, *k.kPi.kind);
  }
  envmap_free(*env_);
  free(env_);
  return res;
}

/*@ requires valid_envmap(env);
  @ requires valid_Type_ptr(t);
  @ requires valid_Kind_ptr(k);
  @ requires valid_string((char*)errorBuffer) && 0 <= errorNum < 2147483647;
  @ requires strlen((char*)errorBuffer) < L_BUFFER_LEN;
  @ assigns errorBuffer[0..L_BUFFER_LEN-1], errorNum;
  @ ensures valid_string((char*)errorBuffer);
  @ ensures \forall size_t n,t; is_allocable{Pre}(n,t) <==> is_allocable{Post}(n,t);
  @ // FIXME: is it really so, the one above?
  @*/
//! 390 398
//! -- | Check if a type expression is of a specified kind
//! -- env |- t : k ?
//! checkType :: MonadError String m => Env -> Type -> Kind -> m ()
//! checkType env t k = do
//!   k' <- inferType env t
//!   if betaEq k k'
//!     then return ()
//!     else reportError ["Actual kind:",show k',
//!                       "\nis not equal to expected:",show k]
bool checkType(const envmap env, Type *t, Kind *k) {
  Kind *k_= inferType(env, t);
  if (k_==NULL) {
    reportError(messages[ENV_MSG_NO_KIND_FOR_TYPE]);
    return false;
  }
  if (betaeq(anyOfKind(k), anyOfKind(k_))) 
    return true;
  else {
    reportError(messages[ENV_MSG_WRONG_ACTUAL_KIND]);
    //    else reportError ["Actual kind:",show k',
    //                      "\nis not equal to expected:",show k]
    //          */
    return false;
  }
}

/*@ requires valid_envmap(env);
  @ requires valid_Type_ptr(t);
  @ requires valid_string((char*)errorBuffer) && 0 <= errorNum < 2147483647;
  @ requires strlen((char*)errorBuffer) < L_BUFFER_LEN;
  @ requires is_allocable(sizeof(Kind),(size_t)1);
  @ assigns errorBuffer[0..L_BUFFER_LEN-1], errorNum;
  @ assigns __allocated[0..MAX_ALLOCS], __size_allocated, __num_allocated;
  @ ensures \at(__num_allocated,Pre)==\at(__num_allocated,Post);
  @ ensures \forall size_t n,t; is_allocable{Pre}(n,t) <==> is_allocable{Post}(n,t);
  @*/
//! 400
//! checkIsType env t = checkType env t Kstar
bool checkIsType(const envmap env, Type *t) {
  Kind *k_KStar = (Kind*)calloc(1, sizeof(Kind));
  if (k_KStar==0) {
    return false;
  }
  //@ assert k_KStar!=\null;
  //@ assert \at(__num_allocated,Pre)+1==__num_allocated;  
  k_KStar->which=eKstar;
  //@ assert \forall size_t n; is_allocable{Pre}(n,t) <==> is_allocable(n-sizeof(Kind),(size_t)(t-1));
  bool res = checkType(env, t, k_KStar);
  //@ assert \at(__num_allocated,Pre)+1==__num_allocated;
  free(k_KStar);
  //@ assert \at(__num_allocated,Pre)==__num_allocated;
  //@ assert \forall size_t n,t; is_allocable{Pre}(n,t) <==> is_allocable(n,t);
  return res;
}

//! 402 422
//! -- | Infer kind of a type expression
//! -- env |- t : ?
//! inferType :: MonadError String m => Env -> Type -> m Kind
//! inferType env (Tvar n) = lookupTvar env n
//! inferType env (Tall "_" t1 t2) = inferType env t2
//! inferType env (Tall n t1 t2) = do
//!   env' <- insertMVar n t1 env
//!   inferType env' t2
//! inferType env (Tapp t m) = do
//!   k <- inferType env t
//!   case k of
//!      Kpi x t1 k1 -> do
//!        checkTerm env m t1
//!        return $ substM x m k1
//!      Kstar -> reportError ["checkType Tapp: expected product kind"]
//! inferType env (Texi n t1 t2) = do
//!   env' <- insertMVar n t1 env
//!   inferType env' t2
//! inferType env (Tand t1 t2) = mapM_ (checkIsType env) [t1,t2] >> return Kstar
//! inferType env (Tor t1 t2) = mapM_ (checkIsType env) [t1,t2] >> return Kstar
//! inferType env Tbot = return Kstar
Kind *inferType(const envmap env, Type *t) {
  Kind *res;
  Kind k_kStar = (Kind) {.which = eKstar, .kStar = {}};
  switch (t->which) {
  case etVar: {
      Kind* knd = envmap_lookupTvar(env, t->tVar.name);
      if (knd==NULL) {
        reportError(messages[ENV_MSG_NO_KIND_FOR_TYPE]);
        return false;
      }
      return copyKindDeep(knd);
  }
  case etAll:
    if (strcmp(t->tAll.name, "_") == 0) 
      return inferType(env, t->tAll.type2);
    else {
      envmap* env_ = envmap_copy(env);
      if (env_ == NULL) {
        reportError(messages[ENV_MSG_IMPOSSIBLE_COPY_ENV]);
        return NULL;
      }
      if (envmap_insertMVar(t->tAll.name,
                            t->tAll.type1, env_) == NULL) {
        reportError(messages[ENV_MSG_CANNOT_INSERT_VARIABLE]);
        return NULL;
      };
      res = inferType(*env_, t->tAll.type2);
      envmap_free(*env_);
      free(env_);
    }
    break;
  case etApp: {
    Kind *k = inferType(env, t->tApp.type);
    if (k==NULL) {
      reportError(messages[ENV_MSG_NO_KIND_FOR_TYPE]);
      return false;
    }
    if (k->which == eKpi) {
      if (checkTerm(env, t->tApp.term, k->kPi.type))
        res = substM_kind(k->kPi.name, t->tApp.term, k->kPi.kind);
      else
        res = NULL;
    } else {
      reportError(messages[ENV_MSG_EXPECTED_PRODUCT_KIND]);
      res = NULL;
    }
    freeKindDeep(k);
    break; }
  case etExi: {
    envmap* env_ = envmap_copy(env);
    if (envmap_insertMVar(t->tExi.name,
                          t->tExi.type1, env_) == NULL) {
      reportError(messages[ENV_MSG_CANNOT_INSERT_VARIABLE]);
      return NULL;
    }
    res = inferType(*env_, t->tExi.type2);
    envmap_free(*env_);
    free(env_);
    break; }
  case etAnd: 
    if (checkIsType(env, t->tAnd.type1) && checkIsType(env, t->tAnd.type2))
      return makeKind(k_kStar);
    else
      return NULL;
  case etOr: 
    if (checkIsType(env, t->tOr.type1) && checkIsType(env, t->tOr.type2))
      return makeKind(k_kStar);
    else
      return NULL;
  case etBot:
    res = makeKind(k_kStar);
    break;
  default:
    reportError(messages[ENV_MSG_UNHANDLED_TYPE]);
    res = NULL;
  }
  return res;
}

/*@ requires valid_envmap(env);
  @ requires valid_Term_ptr(m);
  @ requires valid_Type_ptr(t);
  @ requires valid_string((char*)errorBuffer) && 0 <= errorNum < 2147483647;
  @ requires strlen((char*)errorBuffer) < L_BUFFER_LEN;
  @ assigns errorBuffer[0..L_BUFFER_LEN-1], errorNum;
  @ ensures valid_string((char*)errorBuffer);
  @*/
//! 427 432
//! checkTerm env m t = do
//!   t' <- inferTerm env m
//!   if betaEq t t'
//!     then return ()
//!     else reportError ["Actual type:",show t',
//!                       "\nis not equal to expected:",show t]
bool checkTerm(const envmap env, Term *m, Type *t) {
  Type *t_ = inferTerm(env, m);
  if (t_==NULL) {
    reportError(messages[ENV_MSG_WRONG_ACTUAL_TYPE]);
    return false;
  }
  bool res;
  if (betaeq(anyOfType(t), anyOfType(t_))) 
    res = true;
  else {
    reportError(messages[ENV_MSG_WRONG_ACTUAL_TYPE]);
    //reportError ["Actual type:",show t',
    //                  "\nis not equal to expected:",show t]
    res = false;
  }
  freeTypeDeep(t_);
  return res;
}

/*@ requires valid_envmap(env) && valid_Term_ptr(t) && t->which == emLam;
  @ requires is_allocable(env.bufsize*sizeof(char*)+env.bufsize*sizeof(Any*),(size_t)2);
  @ assigns __allocated[0..MAX_ALLOCS], __size_allocated, __num_allocated;
  @*/
Type *inferLam(const envmap env, Term *t) {
  //@ assert valid_envmap(env);
  //@ assert valid_Term_ptr(t);
  //@ assert valid_Name(t->mLam.name);
  //@ assert valid_Type_ptr(t->mLam.type);
  envmap* env_ = envmap_copy(env);
  //@ assert valid_Term_ptr(t);
  //@ assert valid_Name(t->mLam.name);
  //@ assert valid_Type_ptr(t->mLam.type);
  //@ assert valid_envmap_ptr(&env_);
  if (envmap_insertMVar(t->mLam.name,
                        t->mLam.type, env_) == NULL) {
    reportError(messages[ENV_MSG_CANNOT_INSERT_VARIABLE]);
    return NULL;
  };
  Type *restype = inferTerm(*env_, t->mLam.term);
  if (restype == NULL) {
    reportError(messages[ENV_MSG_IN_INFERTERM_CHECK]);
    return NULL;
  };
  envmap_free(*env_);
  free(env_);
  Name nname = copyName(t->mLam.name);
  Type * type1 = copyTypeDeep(t->mLam.type);
  return makeType((Type) {.which = etAll,
        .tAll = {.name = nname,
                 .type1 = type1,
                 .type2 = restype }
    });
}

/*@ requires valid_envmap(env) && valid_Term_ptr(t) && t->which == emApp;
  @ requires is_allocable(env.bufsize*sizeof(char*)+env.bufsize*sizeof(Any*),(size_t)2);
  @ requires errorBufferValid(0) && errorNum < 2147483646;
  @ assigns __allocated[0..MAX_ALLOCS], __size_allocated, __num_allocated;
  @ assigns errorBuffer[0..4095], errorNum;
  @*/
Type *inferApp(const envmap env, Term *t) {
  //@ assert t->which == emApp && valid_Term_ptr(t->mApp.term1);
  //@ assert \valid(t);
  Type *tf = inferTerm(env, t->mApp.term1);
  if (tf == NULL) {
    reportError(messages[ENV_MSG_IN_INFERTERM]);
    return NULL;
  }
  //@ assert \valid(tf);
  if (tf->which == etAll) {
    //@ assert valid_Term_ptr(t->mApp.term2);
    //@ assert valid_Type_ptr(tf->tAll.type1);
    if (checkTerm(env, t->mApp.term2, tf->tAll.type1)) {
      return substM_type(tf->tAll.name, t->mApp.term2, tf->tAll.type2);
    } else {
      //@ assert errorBufferValid(0) && errorNum < 2147483646;
      reportError(messages[ENV_MSG_IN_INFERTERM]);
      // handler e = reportError ["inferTerm ",show m,"\nWhen checking ",
      //                   show m2,":",show t1,"\n",e,
      //                   "\nEnv",unlines $ map show env]
      return NULL;
    }
  } else {
    reportError(messages[ENV_MSG_NOT_FUNCTION_TYPE]);
    // t -> reportError [show t, "is not a function type"] -- FIXME: better message
    return NULL;
  }
  freeTypeDeep(tf);
  return NULL; // FIXME: is that really so?
}


Type *inferWit(const envmap env, Term *t) {
  if (t->mWit.type->which == etExi) {
    Term *freemeM = substM_term(t->mWit.type->tExi.name,  t->mWit.term1, t->mWit.term2);
    Type *freemeT = substM_type(t->mWit.type->tExi.name,  t->mWit.term1, t->mWit.type->tExi.type2);
    if (checkTerm(env, t->mWit.term1, t->mWit.type->tExi.type1) &&
        checkTerm(env, freemeM, freemeT)) {
      Type *res = makeType((Type) {.which = etExi,
            .tExi = {.name = copyName(t->mWit.type->tExi.name),
                     .type1 = copyTypeDeep(t->mWit.type->tExi.type1),
                     .type2 = copyTypeDeep(t->mWit.type->tExi.type2) }
        });
      freeTermDeep(freemeM);
      freeTypeDeep(freemeT);
      return res;
    } else {
      return NULL;
    }
  } else {
    reportError(messages[ENV_MSG_WITNESS_FOR_NONEXISTENTIAL]);
    return NULL;
  }
}

Type *inferAbs(const envmap env, Term *t) {
  Type* res    = NULL;
  Name  n1  = t->mAbs.name1;
  Name  n2  = t->mAbs.name3;
  Type *t1   = t->mAbs.type2;
  Type *t2   = t->mAbs.type4;
  Term *m1  = t->mAbs.term5;
  Term *m2  = t->mAbs.term6;
  envmap *env_  = envmap_copy(env);
  if (envmap_insertMVar(n1, t1, env_) == NULL) {
    reportError(messages[ENV_MSG_CANNOT_INSERT_VARIABLE]);
    return NULL;
  };
  Type* tt = makeType((Type) {.which = etExi,
        .tExi = { .name = n1,
                  .type1 = t1,
                  .type2 = t2 }
    });
  if (checkTerm(*env_, m1, tt)) {
    if (envmap_insertMVar(n2, t2, env_) == NULL) {
      reportError(messages[ENV_MSG_CANNOT_INSERT_VARIABLE]);
      return NULL;
    }; /* env_ becomes env__ */
    if ((res = inferTerm(*env_, m2)) == NULL) {
      reportError(messages[ENV_MSG_IN]);
      // h2 e = reportError ["In ",show m,"\nIn ",show m2,"\n",e]
    }
  } else {
    reportError(messages[ENV_MSG_IN_INFERTERM_CHECK]);
    // h1 t e = reportError ["inferTerm ",show m,"\nWhen checking ",
    //                      show m1,":",show t,"\n",e]
  }
  envmap_free(*env_);
  free(env_);
  freeType(tt); /* not deep! elements weren't copied */
  return res;
}

Type *inferTup(const envmap env, Term *t) {
  Type* res;
  Type *tmp = nf_type(t->mTup.type);
  if (tmp->which == etAnd) {
    if (checkTerm(env, t->mTup.term1, tmp->tAnd.type1) &&
        checkTerm(env, t->mTup.term2, tmp->tAnd.type2))
      res = copyTypeDeep(t->mTup.type);
    else
      res = NULL;
  } else {
    reportError(messages[ENV_MSG_TUPLE_NOT_AND]);
    res = NULL;
  }
  freeTypeDeep(tmp);
  return res;
}

Type *inferPi(const envmap env, Term *t, Term* prod) {
  Type* res;
  Type *tmp = inferTerm(env, prod);
  if (tmp == NULL) {
    reportError(messages[ENV_MSG_IN_INFERTERM]);
    return NULL;
  }
  if (tmp->which == etAnd) {
    if (t->which == emPi1) {
      res = copyTypeDeep(tmp->tAnd.type1);
    } else {
      res = copyTypeDeep(tmp->tAnd.type2);
    }
  } else {
    res = NULL;
    reportError(messages[ENV_MSG_IS_NOT_PAIR]);
    // reportError [show t, "is not a pair type"]
  }
  freeTypeDeep(tmp);
  return res;
}

Type *inferIn(const envmap env, Term *t, Type *inType, Term* bTerm, Type *bType) {
  Type* res;
  if (inType->which == etOr) {
    if (checkTerm(env, bTerm, bType)) 
      res = copyTypeDeep(inType);
    else
      res = NULL;
  } else {
    // reportError ["in1 into",show t,"which is not disjunction"]
    reportError(messages[ENV_MSG_IN_CASE]);
    res = NULL;
  }
  return res; // FIXME: is that really so?
}

Type *inferCas(const envmap env, Term *t) {
  Type* res;
  Type *tmp = makeType((Type) {.which = etOr, 
        .tOr = { .type1 = t->mCas.casIn1.type, 
                 .type2 = t->mCas.casIn2.type }
    });
  
  res = NULL;
  if (checkTerm(env, t->mCas.term, tmp)) {
    envmap* env1 = envmap_copy(env);
    if (envmap_insertMVar(t->mCas.casIn1.name,
                          t->mCas.casIn1.type, env1) == NULL) {
      reportError(messages[ENV_MSG_CANNOT_INSERT_VARIABLE]);
      return NULL;
    };
    Type *t3 = inferTerm(*env1, t->mCas.casIn1.term);
    if (t3 == NULL) {
      reportError(messages[ENV_MSG_IN_INFERTERM]);
      return NULL;
    }
    envmap* env2 = envmap_copy(env);
    if (envmap_insertMVar(t->mCas.casIn2.name,
                          t->mCas.casIn2.type, env2) == NULL) {
      reportError(messages[ENV_MSG_CANNOT_INSERT_VARIABLE]);
      return NULL;
    };
    if (checkTerm(*env2, t->mCas.casIn2.term, t3))
      res = t3;
    else
      freeTypeDeep(t3);
    envmap_free(*env1);
    envmap_free(*env2);
    free(env1);
    free(env2);
  }
  freeType(tmp); /* not deep! elements weren't copied */
  return res;
}

Type *inferEps(const envmap env, Term *t) {
  Type* res;
  Type *bot = makeType((Type) {.which = etBot,
        .tBot = {}
    });
  
  res = checkTerm(env, t->mEps.term, bot) ? t->mEps.type : NULL;
  freeType(bot);
  return res;
}

/*@ requires valid_envmap(env);
  @ requires valid_Term_ptr(t);
  @ assigns \nothing;
  @ ensures valid_Type_ptr(\result);
  @*/
//! 434 517
//! inferTerm :: MonadError String m => Env -> Term -> m Type
//! inferTerm env (Mvar n) = lookupMvar env n
//! inferTerm env (Mlam n t m) = do
//!   env' <- insertMVar n t env
//!   t' <- inferTerm env' m
//!   return (Tall n t t')
//! 
//! inferTerm env m@(Mapp m1 m2) = do
//!   tf <- inferTerm env m1
//!   case tf of
//!     Tall n t1 t2 -> do
//!       checkTerm env m2 t1 `catchError` handler
//!       return $ substM n m2 t2
//!         where
//!           handler e = reportError ["inferTerm ",show m,"\nWhen checking ",
//!                             show m2,":",show t1,"\n",e,
//!                             "\nEnv",unlines $ map show env]
//! 
//!     t -> reportError [show t, "is not a function type"] -- FIXME: better message
//!      
//! -- [m1,m2]_{∃ x:phi1.phi2}
//! inferTerm env (Mwit (Texi n t1 t2)  m1 m2) = do
//!   checkTerm env m1 t1
//!   checkTerm env (substM n m1 m2) (substM n m1 t2)
//!   return (Texi n t1 t2)
//! 
//! inferTerm env (Mwit t  m1 m2) =
//!   reportError ["Witness for a nonexistential type"]
//! 
//! -- abstract <n1:t1,n2:t2> = m1 in m2
//! inferTerm env m@(Mabs n1 t1 n2 t2 m1 m2) = do
//!   env' <- insertMVar n1 t1 env
//!   let t = (Texi n1 t1 t2)
//!   checkTerm env' m1 t `catchError` (h1 t)
//!   env'' <- insertMVar n2 t2 env'
//!   inferTerm env'' m2 `catchError` h2
//!      where
//!        h1 t e = reportError ["inferTerm ",show m,"\nWhen checking ",
//!                             show m1,":",show t,"\n",e]
//!        h2 e = reportError ["In ",show m,"\nIn ",show m2,"\n",e]
//! 
//! inferTerm env (Mtup t m1 m2)
//!   | (Tand t1 t2) <- nf t = do
//!     checkTerm env m1 t1
//!     checkTerm env m2 t2
//!     return t
//!   | otherwise = reportError ["Tuple not of and type"]
//! 
//! inferTerm env (Mpi1 m) = do
//!   t <- inferTerm env m
//!   case t of
//!     (Tand t1 t2) -> return t1
//!     _ -> reportError [show t, "is not a pair type"]
//! 
//! inferTerm env (Mpi2 m) = do
//!   t <- inferTerm env m
//!   case t of
//!     (Tand t1 t2) -> return t2
//!     _ -> reportError [show t, "is not a pair type"]
//! 
//! inferTerm env (Min1 t m)
//!   | Tor t1 t2 <- t = do
//!     checkTerm env m t1
//!     return t
//!   | otherwise = reportError ["in1 into",show t,"which is not disjunction"]
//! 
//! inferTerm env (Min2 t m)
//!   | Tor t1 t2 <- t = do
//!     checkTerm env m t2
//!     return t
//!   | otherwise = reportError ["in2 into",show t,"which is not disjunction"]
//! 
//! inferTerm env (Mcas m (n1,t1,m1) (n2,t2,m2)) = do
//!   checkTerm env m (Tor t1 t2)
//!   env1 <- insertMVar n1 t1 env
//!   t3 <- inferTerm env1 m1
//!   env2 <- insertMVar n2 t2 env
//!   checkTerm env2 m2 t3
//!   return t3
//! 
//! inferTerm env (Meps t m) = do
//!   checkTerm env m Tbot
//!   return t
//! 
Type *inferTerm(const envmap env, Term *t) {
  Type *res;
  switch (t->which) {
  case emVar: {
      Type* typ = envmap_lookupMvar(env, t->mVar.name);
      if (typ==NULL) {
        reportError(messages[ENV_MSG_UNDECLARED_MVAR]);
        return NULL;
      }
      return copyTypeDeep(typ);}
  case emLam:
    res = inferLam(env, t);
    break;
  case emApp:
    res = inferApp(env, t);
    break;
  case emWit:
    res = inferWit(env, t);
    break;
  case emAbs:
    res = inferAbs(env, t);
    break;
  case emTup:
    res = inferTup(env, t);
    break;
  case emPi1:
    res = inferPi(env, t, t->mPi1.term);
    break;
  case emPi2:
    res = inferPi(env, t, t->mPi2.term);
    break;
  case emIn1:
    res = inferIn(env, t, t->mIn1.type, t->mIn1.term, t->mIn1.type->tOr.type1);
    break;
  case emIn2:
    res = inferIn(env, t, t->mIn2.type, t->mIn2.term, t->mIn2.type->tOr.type2);
    break;
  case emCas:
    res = inferCas(env, t);
    break;
  case emEps:
    res = inferEps(env, t);
    break;
  default:
    reportError(messages[ENV_MSG_UNHANDLED_TERM]);
    res = NULL;
  }
  return res;
}
