/****************************************************************************
 * Copyright (C) 2016 by University of Warsaw                               *
 *                                                                          *
 * This file is part of IntuitionTC.                                        *
 *                                                                          *
 ****************************************************************************/
/**
 * @file vars_set.h
 * @brief Finite sets with union operation.
 * @author Aleksy Schubert
 */
#ifndef _VARS_SET_H
#define _VARS_SET_H  1

#include <string.h>
#include "checker.h"

#define INITIAL_SIZE 1024
//@ logic size_t L_INITIAL_SIZE = (size_t)1024;

/** @brief The representation of the set that holds strings.
 * 
 *  The non-null entries in the #vals array represent strings that belong
 *  to the set.
 */
typedef struct nameSet {
  /** @brief The size of the allocated area in which set elements can 
   *   be stored. */ 
  int     size;
  /** @brief The array in which the elements of the set can be stored. */ 
  Name*   vals;
} nameSet;


/*@ predicate positive_size{L}(nameSet* s) = s->size > 0; @*/
/*@ predicate valid_vals{L}(nameSet* s) = s->size>=0 ==> is_allocated_size((char*)(s->vals),(size_t)s->size);  @*/
/*@ predicate vals_unique{L}(nameSet* s) = \forall integer i,j; 
  @   0 <= i < j < s->size && s->vals != \null && 
  @   s->vals[i]!=\null && s->vals[j]!=\null ==>
  @     strcmp(s->vals[i],s->vals[j])!=0;
  @*/ 

/*@ predicate valid_vals_valid{L}(nameSet* s) = \forall integer i; 
  @   0 <= i < s->size && s->vals[i] != \null  ==> 
  @              (is_allocated_size(s->vals[i],(size_t)\block_length(s->vals[i])) &&
  @              0 <= \block_length(s->vals[i]) <= UINT32_MAX &&
  @              valid_string(s->vals[i]));
  @*/

/*@ predicate well_separated{L}(nameSet* s) = 
  @  \separated(s, s->vals+(0..s->size-1)) &&
  @  \separated(s, { s->vals[i]  | 
  @             integer i; 0 <= i < s->size && s->vals[i]!=0 }) &&
  @  \separated(s->vals, { (s->vals)[i]  | 
  @             integer i; 0 <= i < s->size && s->vals[i]!=0 }) &&
  @  \separated(s, __allocated+(0..MAX_ALLOCS)) &&
  @  \separated(s->vals+(0..s->size-1), __allocated+(0..MAX_ALLOCS)) &&
  @  \forall integer i; 0 <= i < s->size && s->vals[i] != \null ==>
  @    \separated(__allocated+(0..MAX_ALLOCS),
  @               s->vals[i]+(0..\block_length(s->vals[i])-1));
  @*/

/*@ predicate valid_nameSet{L}(nameSet* s) = \valid(s) &&
  @   is_allocated((char*)s) &&
  @   positive_size(s) && valid_vals(s) && vals_unique(s) && 
  @   valid_vals_valid(s) && well_separated(s);
  @*/

/*@ predicate fresh_Set{L}(nameSet* s) = \valid(s) && is_allocated((char*)s) &&
  @   positive_size(s) && valid_vals(s) && vals_unique(s) && 
  @   valid_vals_valid(s) && well_separated(s);
  @*/



nameSet* unionSet(nameSet* s1, nameSet* s2);

/* union set and call freeSetShallow on operands */
nameSet* unionSetFreeShallow(nameSet* s1, nameSet* s2);
Name minusSet(nameSet* s, Name elem);
/* s is modified! */
nameSet* addSet(nameSet* s, Name elem);
nameSet* createSet();
nameSet* singletonSet(Name elem);
void freeSetShallow(nameSet* s);
void freeSetDeep(nameSet* s);
/* returns index or -1 */
int isIn(nameSet* s, Name elem);


#endif
