/****************************************************************************
 * Copyright (C) 2016 by University of Warsaw                               *
 *                                                                          *
 * This file is part of IntuitionTC.                                        *
 *                                                                          *
 ****************************************************************************/

/**
 * @file main.h
 * @brief The main function, analysis of command line parameters and 
 *        calling appropriate logic.
 * @author Aleksy Schubert
 */

/** @brief The variable is used to communicate errors to functions that
 * do not have appropriate arguments.
 *
 * Initially it is equal to 0.
 */
extern int waserror;

/** @brief The global variable that contains the proof parsing status.
 * 
 * It serves to communicate this status to the yacc parser code.
 * It has value 1 when the proof should be parsed from the input file and
 * it has value 0 otherwise. Initially it is equal to 0.
 */
extern int proof_parsing;


/** @brief The procedure prints the information about the syntax error in the
 *  TPTP format.
 *
 *  @param s the string passed from the yacc parser with information on
 *           the reason of error
 */
void yyerror(char* s);

/** @brief The procedure prints out information on how to invoke the program.
 *
 * @param pname the name of the executable the program is started from
 */
void printUsage(char* pname);

/** @brief The procedure handles the option of syntax checking for
 *         input formulas.
 *
 *  @param filename the name of the file to parse formulas from
 *  @return 0 in case all is correctly parsed, other values on error
 */
int formulas(char* filename);

/** @brief The procedure handles the option of typechecking of proofs
 *         for given formulas.
 *
 *  @param filename the name of the file to parse a proof from
 *  @return 0 in case all is correctly parsed and typeckecked, 
 *          other values on error
 */
int proofs(char* filename);

/** @brief The main procedure of the program.
 *
 *  The procedure parses the command line options with getopt and
 *  directs control to appropriate functions resulting from
 *  command line options. In case options are improper usage of
 *  the program is printed out.
 *
 *  @param argc the number of command line chunks
 *  @param argv the array of strings that contain command line chunks
 *  @return it returns the exit status of the program
 */
int main(int argc, char *argv[]);

/** @brief Typechecks the given term as the proof for the given type.
 *
 * The procedure takes the given term, creates appropriate environment
 * with the carrier type and types for the free predicates used in the
 * formula.
 *
 * @param term the term to typecheck
 * @param type the type i.e. formula to check if the term is its proof
 * @return true when typechecking is successful, false on error or when
 *         the proof is not for the given formula
 */
int typecheckFormula(Term* term, Type* type);

/** @brief The procedure is used to add to environment types of free
 *         predicates that occur in the formula.
 *
 * @param type the type to search for free predicates
 * @param env the environment to insert types for free predicates to
 * @param arity the computed so far arity of the next free predicate to insert
 * @return 0 on error, other values when update is successful
 */
int updateTypeVariables(Type* type, envmap *env, int arity);

/** @brief The procedure is used to add to environment types of free
 *         predicates that occur in the given case subexpression of a
 *         term.
 *
 * @param cas the cas subexpression to search for free predicates
 * @param env the environment to insert types for free predicates to
 * @return 0 on error, other values when update is successful
 */
int updateTypeVariablesCas(McasIn cas, envmap *env);

/** @brief The procedure is used to add to environment types of free
 *         predicates that occur in the given term.
 *
 * @param term the term to search for free predicates
 * @param env the environment to insert types for free predicates to
 * @return 0 on error, other values when update is successful
 */
int updateTypeVariablesTerm(Term * term, envmap *env);

/** @brief The procedure creates the Kind that represents the given arity
 *         of a predicate. 
 *
 * @param arity the number of arguments of the standard carrier type
 * @return the pointer to the freshly created Kind that represents the
 *         arity or @c NULL on error
 * 
 */
Kind* createArity(int arity);


