/****************************************************************************
 * Copyright (C) 2016 by University of Warsaw                               *
 *                                                                          *
 * This file is part of IntuitionTC.                                        *
 *                                                                          *
 ****************************************************************************/
#ifndef _PARSERHELPERS_H
#define _PARSERHELPERS_H
/**
 * @file parserHelpers.h
 * @brief Functions that help making yacc file concise.
 * @author Aleksy Schubert
 */
#include "envmap_base.h"
#include "vars_set.h"
#include "checker.h"

/** @brief The type of the values located in the first field of the
 *  union #YYSTYPE passed as the return value in the parser.
 *
 *  In some cases only this value is used to inform on which
 *  symbol has been encountered by the parser.
 *  The first 5 values match the ones of the enum type #eAny.
 */
typedef enum {
  /** The value contains #Type structure */
  eAnyTypeH,
  /** The value contains #Term structure */
  eAnyTermH,
  /** The value contains #Kind structure */
  eAnyKindH,
  /** The value contains ::Name */
  eAnyNameH,
  /** The value contains #Bind structure */
  eAnyBindH,
  /** Information that equivalence symbol was encountered */
  eConnDoubleArrow,
  /** Information that implication symbol was encountered */
  eConnRightArrow,
  /** Information that reverse implication symbol was encountered */
  eConnLeftArrow,
  /** Information that disjunction symbol was encountered */
  eConnOr,
  /** Information that XOR symbol was encountered */
  eConnXor,
  /** Information that NOR symbol was encountered */
  eConnNor,
  /** Information that NAND symbol was encountered */
  eConnNand,
  /** Information that negation symbol was encountered */
  eConnNot,
  /** Information that universal quantifier symbol was encountered */
  eConnForall,
  /** Information that existential quantifier symbol was encountered */
  eConnExists,
  /** Information that unhandled symbol was encountered */
  eConnUnhandled 
} eAnyH;

/** @brief A wrapper structure to pass along identifier names within the parser */
typedef struct string {
  /** @brief The type field; should always contain the value 0. */
  eAnyH which;
  /** @brief The actual location of the identifier. */
  Name name;
} string;


#ifndef YYSTYPE_IS_DECLARED
/** @brief The guard to make sure the union #YYSTYPE is declared only once */
#define YYSTYPE_IS_DECLARED 1
/** @brief The union type of the values passed as the return value in
    the parser.
 */
typedef union {
  /** @brief Information on which particular type constructor has been read. */
  eAnyH   kindof;
  /** @brief Container with a parsed subexpression. */
  Any     *any;
  /** @brief A name of a parsed identifier to be used in terms. */
  string  name;
  /** @brief A #nameSet of terms or types used to construct terms that involve
   *  sequences of names (e.g. quantified formulas). */
  nameSet *ns;
  /** @brief An #envmap environment that contains a number of formulas or terms
   *  that involve sequences of terms (e.g. predicate applications). */
  envmap  *emap;
  /** @brief A structure to hold a case of a case term. */
  McasIn  *caseterm;
} YYSTYPE;
#endif /* !YYSTYPE_IS_DECLARED */

#include "y.tab.h"


/**
 * @brief The procedure parses the content of the file opened for parsing.
 *
 * The code of the function is automatically generated by yacc and flex
 * from syntaxAnalyser.y and lexicalAnalyzer.l files.
 *
 * @return the parsing status
 **/
int yyparse(void);

/**
 * @brief The procedure creates a basic binary operator such as AND or
 * OR.
 *
 * It allocates fresh memory for the new #Type to represent
 * the operator as well as #Any container for the type.
 * The arguments of the type are retrieved from arg1 and
 * arg2 #Any arguments. It does not check if the arguments
 * are well formed, but it checks if local operations
 * (allocation and filling the arguments) are correct.
 *
 * @param kind  the kind of the binary operator to create
 * @param arg1  the first argument of the operator
 * @param arg2  the second argument of the operator
 *
 * @return a new #Any container for the new type or @c NULL in
 *   case of error, it emits error message in case of error
 **/
Any* createBasicBinary(eType kind, Any* arg1, Any* arg2);

/**
 * @brief The procedure creates a string to be used as identifier.
 *
 * It allocates fresh memory for the new ::Name and copies there
 * the content of the lexing global variable #yytext.
 *
 * @param res  the #YYSTYPE union to store the pointer to the
 *             identifier string to
 *
 * @return the newly allocated copy of the string obtained from lexing or
 *   @c NULL in case of allocation error
 **/
Name createWord(YYSTYPE* res);

/**
 * @brief The procedure composes a new type from the arguments and the kind
 * of the type to be created. 
 *
 * It checks if the arguments are indeed types. It handles currently only
 * AND and OR types. All other kinds of types are not handled here and
 * treated as errors.
 *
 * @param tp      the kind of type to be created
 * @param first   the first argument of the newly created type
 * @param second  the second argument of the newly created type
 *
 * @return the #Any structure that holds the type created in the procedure
 *   or @c NULL in case of error
 **/
Any* composeAnyType(eType tp, Any* first, Any* second);

/**
 * @brief The procedure composes a new type from the arguments and the
 *        information on the expected result contained in the #YYSTYPE
 *        union in the first argument.
 *
 * It checks if the second and third arguments are indeed types. It
 * handles currently: left arrow, right arrow, double arrow, NOR, NAND, OR.
 * All other kinds of types, especially those that do not represent
 * connectives or represent unary ones, are not handled here and treated as
 * errors.
 *
 * @param res     the structure for the expected return type
 * @param arg1    the first argument of the newly created type
 * @param arg2    the second argument of the newly created type
 *
 * @return the #Any structure that holds the type created in the procedure
 *   or @c NULL in case of error
 **/
Any* composeBinaryConnective(YYSTYPE res, Any* arg1, Any* arg2);

/**
 * @brief The procedure composes a new type from the argument and
 *        unary connective prescribed in the #YYSTYPE union in the
 *        first argument.
 *
 * It checks if the second argument is indeed a type. It handles currently
 * negation only. All other kinds of types, especially those that do not
 * represent connectives or represent binary ones, are not handled here and
 * treated as errors.
 *
 * @param res     the structure for the expected return type
 * @param arg     the argument of the newly created type
 *
 * @return the #Any structure that holds the type created in the procedure
 *   or @c NULL in case of error
 **/
Any* composeUnaryConnective(YYSTYPE res, Any* arg);

/**
 * @brief The procedure frees memory of a quantifier chain allocated
 * in #composeQuantifier.
 *
 * @param start  pointer to chain of quantifiers
 **/
void freeQChain(Type* start);

/**
 * @brief The procedure is used to create a formula that starts with
 * either universal or existential quantifier.
 *
 * It checks if the given @c formula is accessible and then gradually 
 * creates a series of formulas with bigger and bigger number of initial
 * quantifiers. It uses the strings in @c ns for the variable names to
 * quantify for. In case @c tp is not of an appropriate type it
 * generates error.
 *
 * @param tp       the variety of the quantifier formula to create
 * @param ns       the set of names for the quantified variables
 * @param formula  the #Type that constitutes the body of the formula
 *
 * @return the #Any structure that contains the type with the quantified
 *   formula inside or @c NULL in case of error
 **/
Any* composeQuantifier(eAnyH tp, nameSet* ns, Any* formula);


/**
 * @brief The procedure is used to create #Term in #Any wrapper with variable
 *        the name of which is given in the argument #string.
 * 
 * The memory for the new representation is freshly allocated here and
 * the content of the #string is copied there.
 *
 * @param arg  the #string with name of the variable
 *
 * @return the newly allocated #Any structure with the new variable term in
 *   its contents or @c NULL on error
 **/
Any* composeTermVariable(string arg);

/**
 * @brief The procedure is used to create #Type in #Any wrapper with variable
 *        the name of which is given in the argument #string.
 *
 * The memory for the new representation is freshly allocated here and
 * the content of the #string is copied there.
 *
 * @param arg  the string with name of the variable
 *
 * @return the newly allocated #Any structure with the new type variable in
 *   its contents or @c NULL on error
 **/
Any* composeTypeVariable(string arg);

/**
 * @brief The procedure creates a #Type that is a predicate variable
 *        applied to the given sequence of @c arguments.
 * 
 * The function creates a sequence of applications that starts with
 * the type variable the name of which is given in @c name parameter
 * and arguments of which are stored in the @c arguments parameter.
 * It emits a message and returns @c NULL on error.
 *
 * @param name the #string with the name of the predicate variable
 *        that is to be applied to the given arguments
 * @param arguments #envmap with arguments to the predicate variable
 * @return the newly allocated #Any structure with the new #Type in
 *   its contents or @c NULL on error
 */
Any* composeFunctorWithArguments(string name, envmap *arguments);

/**
 *  @brief The procedure translates given #Type to #Term if this is possible.
 *
 *  The procedure checks if the argument is properly formed and wraps the
 *  result in #Any structure. The actual work is done in typeToTermRaw().
 *  The procedure works for a type variable applied to its arguments. Some
 *  parsing within the TPTP parser is common for both types and terms and
 *  we have to translate one to the other when the initial assumption that
 *  the construct is a type turns out to be wrong.
 *
 *  @param mtype the #Type to convert to #Term
 *
 *  @return the resulting #Term wrapped in #Any structure or @c NULL on
 *          error
 *
 */
Any* typeToTerm(Any* mtype);

/**
 *  @brief The procedure translates given #Type to #Term if this is possible.
 *
 *  Some parsing within the TPTP parser is common for both types and
 *  terms and we have to translate one to the other when the initial
 *  assumption that the construct is a type turns out to be wrong.
 *
 *  @param mtype the #Type to convert to #Term
 *
 *  @return the resulting #Term wrapped in #Any structure or @c NULL on
 *          error
 *
 */
Term* typeToTermRaw(Type* mtype);


/**
 * @brief The procedure creates the <em>ex falso quod libet</em> proof
 *        term from given #Term and #Type.
 *
 * The procedure checks shallowly if the arguments are correctly formed
 * and creates <em>ex falso quod libet</em> proof term which is
 * subsequently wrapped in an #Any structure. Note that we do not
 * check the logical structure here.  In particular @c term parameter
 * need not be a proof of falsity here.
 *
 * @param term  a #Term that is turned into <em>ex falso proof</em>
 * @param formula  the #Type with the resulting formula
 *
 * @result #Any structure that contains the resulting <em>ex falso</em>
 *         term or @c NULL on error
 * 
 */
Any* createExFalso(Any* term, Any* formula);

/**
 * @brief The procedure creates the <em>lambda</em> proof term
 *        from given #Name of the argument, its #Type and body #Term.
 *
 * The procedure checks shallowly if the arguments are correctly formed
 * and creates <em>lambda term</em> proof which is subsequently
 * wrapped in an #Any structure. Note that we do not check the logical
 * structure here.
 *
 * @param vname    the #string with the name of the argument variable
 * @param vtype    the #Type of the argument
 * @param term     the #Term with the body of the <em>lambda term</em>
 *
 * @result #Any structure that contains the resulting <em>lambda</em> term
 *         or @c NULL on error
 */
Any* createLambda(string vname, Any* vtype, Any* term);

/**
 * @brief The procedure creates the <em>application</em> proof term
 *        from given two #Term terms.
 *
 * The procedure checks shallowly if the arguments are correctly formed
 * and creates <em>application term</em> proof which is subsequently
 * wrapped in an #Any structure. Note that we do not check the logical
 * structure here.
 *
 * @param term1    the operand term
 * @param term2    the argument term
 *
 * @result #Any structure that contains the resulting <em>application</em> term
 *         or @c NULL on error
 */
Any* createApp(Any* term1, Any* term2);

/**
 * @brief The procedure creates the <em>abstraction</em> (i.e. existential
 *        quantifier elimination) proof term from given two names of fresh
 *        variables, their types, implementation term and usage term.
 *
 * The procedure checks shallowly if the arguments are correctly formed
 * and creates <em>abstraction term</em> proof (i.e. existential quantifier
 * elimination term) which is subsequently wrapped in an #Any structure.
 * Note that we do not check the logical structure here.
 *
 * @param name1    the #string with the name of the abstracted object
 * @param type2    the type of the abstracted object
 * @param name3    the #string with the name of the abstracted implementation
 * @param type4    the type of the abstracted implementation
 * @param term5    the term with the abstracted implementation
 * @param term6    the term which used the abstracted implementation
 *
 * @result #Any structure that contains the resulting <em>abstraction</em> term
 *         or @c NULL on error
 */
Any* createAbstract(string name1, Any* type2, string name3,
                    Any* type4, Any* term5, Any* term6);

/**
 * @brief The procedure creates the <em>exists</em> (i.e. existential
 *        quantifier introduction) proof term from given result type,
 *        the witness term and the implementation term.
 *
 * The procedure checks shallowly if the arguments are correctly formed
 * and creates <em>exists term</em> proof (i.e. existential quantifier
 * introduction term) which is subsequently wrapped in an #Any structure.
 * Note that we do not check the logical structure here.
 *
 * @param type    the existential type of the resulting term
 * @param term1   the witness term of the existential quantifier introduction
 * @param term1   the implementation term of the existential quantifier
 *                introduction
 *
 * @result #Any structure that contains the resulting <em>exists</em> term
 *         or @c NULL on error
 */
Any* createExists(Any* type, Any* term1, Any* term2);

/**
 * @brief The procedure creates the <em>case</em> (i.e. alternative
 *        elimination) proof term from given implementation term and
 *        cases.
 *
 * The procedure checks shallowly if the arguments are correctly formed
 * and creates <em>case term</em> proof (i.e. alternative elimination
 * term) which is subsequently wrapped in an #Any structure.
 * Note that we do not check the logical structure here.
 *
 * @param term    the implementation term of the alternative elimination
 * @param case1   the case expression to handle the first option of the
 *                alternative
 * @param case2   the case expression to handle the second option of the
 *                alternative
 *
 * @result #Any structure that contains the resulting <em>case</em> term
 *         or @c NULL on error
 */
Any* createCase(Any* term, McasIn* case1, McasIn* case2);

/**
 * @brief The procedure creates the <em>injection</em> (i.e. alternative
 *        introduction) proof term from given implementation and the
 *        resulting type.
 *
 * The procedure checks shallowly if the arguments are correctly formed
 * and creates <em>injection term</em> proof (i.e. alternative introduction
 * term) which is subsequently wrapped in an #Any structure.
 * Note that we do not check the logical structure here.
 *
 * @param term    the implementation term of the alternative introduction
 * @param type    the type of the resulting term
 * @param which   information on which of the two injections is created
 *                (it is equal to either 1 or 2)
 *
 * @result #Any structure that contains the resulting <em>injection</em> term
 *         or @c NULL on error
 */
Any* createIn(Any* term, Any* type, int which);

/**
 * @brief The procedure creates the <em>projection</em> (i.e. conjunction
 *        elimination) proof term from given product term.
 *
 * The procedure checks shallowly if the arguments are correctly formed
 * and creates <em>projection term</em> proof (i.e. conjunction elimination
 * term) which is subsequently wrapped in an #Any structure.
 * Note that we do not check the logical structure here.
 *
 * @param term    the product term of the conjunction elimination
 * @param which   information on which of the two projections is created
 *                (it is equal to either 1 or 2)
 *
 * @result #Any structure that contains the resulting <em>projection</em> term
 *         or @c NULL on error
 */
Any* createProj(Any* term, int which);

/**
 * @brief The procedure creates the <em>tuple</em> (i.e. conjunction
 *        introduction) proof term from given two product terms and
 *        type of the result.
 *
 * The procedure checks shallowly if the arguments are correctly formed
 * and creates <em>tuple term</em> proof (i.e. conjunction introduction
 * term) which is subsequently wrapped in an #Any structure.
 * Note that we do not check the logical structure here.
 *
 * @param type    the type of the result
 * @param term1   the term that becomes the first coordinate of the tuple
 * @param term2   the term that becomes the second coordinate of the tuple
 *
 * @result #Any structure that contains the resulting <em>tuple</em> term
 *         or @c NULL on error
 */
Any* createTuple(Any* type, Any* term1, Any* term2);


/**
 * @brief The procedure creates the <em>term variable</em> proof term
 *        name of the variable.
 *
 * The procedure creates <em>term variable</em> proof (i.e. axiom
 * proof term) which is subsequently wrapped in an #Any structure.
 * Note that we do not check the logical structure here.
 *
 * @param vname    the name of the term variable
 *
 * @result #Any structure that contains the resulting <em>variable</em> term
 *         or @c NULL on error
 */
Any* createTermVar(string vname);

/**
 * @brief The procedure creates the <em>case</em> subexpression from given
 *        name of the variable, its type and the implementation term.
 *
 * The procedure checks shallowly if the arguments are correctly formed
 * and creates <em>case</em> subexpression that is subsequently used
 * in creation of <em>case</em> proofs (i.e. alternative elimination
 * proof terms).
 * Note that we do not check the logical structure here.
 *
 * @param name    the name of the case analysed term when it matches
 *                the current case
 * @param term1   the type of the current case
 * @param term2   the term that realises handling of the current case
 *
 * @result #McasIn structure that contains the resulting <em>case</em> 
 *         subexpression or @c NULL on error
 */
McasIn* createCaseTerm(string name, Any* type, Any* term);

/** @brief The procedure creates a name that does not occur in the given
 *         environment.
 *
 *  @param env  the environment to guarantee freshness against
 *
 *  @result the name of the fresh variable
 */
Name getFreshName(envmap *env);

#endif
